# Copyright 2016 NTT DATA
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for host failure TaskFlow
"""
import copy
from unittest import mock

import ddt

from masakari.compute import nova
from masakari import conf
from masakari import context
from masakari.engine.drivers.taskflow import host_failure
from masakari.engine import manager
from masakari import exception
from masakari import test
from masakari.tests.unit import fakes

CONF = conf.CONF


@ddt.ddt
@mock.patch.object(nova.API, "enable_disable_service")
@mock.patch.object(nova.API, "lock_server")
@mock.patch.object(nova.API, "unlock_server")
class HostFailureTestCase(test.TestCase):

    def setUp(self):
        super(HostFailureTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        # overriding 'wait_period_after_evacuation' and
        # 'wait_period_after_service_update' to 2 seconds to
        # reduce the wait period.
        self.override_config("wait_period_after_evacuation", 2)
        self.override_config("wait_period_after_service_update", 2)
        self.override_config("evacuate_all_instances",
                             False, "host_failure")
        self.instance_host = "fake-host"
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()

    def _verify_instance_evacuated(self, old_instance_list):
        for server in old_instance_list:
            instance = self.novaclient.get_server(self.ctxt, server)

            if getattr(instance, 'OS-EXT-STS:vm_state') in \
                    ['active', 'stopped', 'error']:
                self.assertIn(getattr(instance, 'OS-EXT-STS:vm_state'),
                              ['active', 'stopped', 'error'])
            else:
                if getattr(instance, 'OS-EXT-STS:vm_state') == 'resized' and \
                        getattr(instance, 'OS-EXT-STS:power_state') != 4:
                    self.assertEqual('active',
                                     getattr(instance, 'OS-EXT-STS:vm_state'))
                else:
                    self.assertEqual('stopped',
                                     getattr(instance, 'OS-EXT-STS:vm_state'))

            if CONF.host_failure.ignore_instances_in_error_state and getattr(
                    instance, 'OS-EXT-STS:vm_state') == 'error':
                self.assertEqual(
                    self.instance_host, getattr(
                        instance, 'OS-EXT-SRV-ATTR:hypervisor_hostname'))
            else:
                self.assertNotEqual(
                    self.instance_host, getattr(
                        instance, 'OS-EXT-SRV-ATTR:hypervisor_hostname'))

    def _test_disable_compute_service(self, mock_enable_disable):
        task = host_failure.DisableComputeServiceTask(self.ctxt,
                                                      self.novaclient)
        task.execute(self.instance_host)

        mock_enable_disable.assert_called_once_with(
            self.ctxt, self.instance_host)

    def _test_instance_list(self, instances_evacuation_count):
        task = host_failure.PrepareHAEnabledInstancesTask(self.ctxt,
                                                          self.novaclient)
        instances = task.execute(self.instance_host)
        instance_uuid_list = []
        for instance_id in instances['instance_list']:
            instance = self.novaclient.get_server(self.ctxt, instance_id)
            if CONF.host_failure.ignore_instances_in_error_state:
                self.assertNotEqual("error",
                                    getattr(instance, "OS-EXT-STS:vm_state"))
            if not CONF.host_failure.evacuate_all_instances:
                ha_enabled_key = (CONF.host_failure
                                      .ha_enabled_instance_metadata_key)
                self.assertTrue(instance.metadata.get(ha_enabled_key, False))

            instance_uuid_list.append(instance.id)

        self.assertEqual(instances_evacuation_count,
                         len(instances['instance_list']))

        return {
            "instance_list": instance_uuid_list,
        }

    def _evacuate_instances(self, instance_list, mock_enable_disable,
                            reserved_host=None):
        task = host_failure.EvacuateInstancesTask(
            self.ctxt, self.novaclient,
            update_host_method=manager.update_host_method)
        old_instance_list = copy.deepcopy(instance_list['instance_list'])

        if reserved_host:
            task.execute(self.instance_host,
                         instance_list['instance_list'],
                         reserved_host=reserved_host)

            self.assertTrue(mock_enable_disable.called)
        else:
            task.execute(
                self.instance_host, instance_list['instance_list'])

        # make sure instance is active and has different host
        self._verify_instance_evacuated(old_instance_list)

    @mock.patch.object(nova.API, "get_server")
    @mock.patch.object(nova.API, "evacuate_instance")
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.LOG')
    def test_instance_evacute_error(self, _mock_log, _mock_evacuate,
            _mock_get, mock_unlock, mock_lock, mock_enable_disable):
        task = host_failure.EvacuateInstancesTask(
            self.ctxt, self.novaclient,
            update_host_method=manager.update_host_method)

        def get_fake_server(server, status):
            fake_server = copy.deepcopy(server)
            setattr(fake_server, 'OS-EXT-SRV-ATTR:hypervisor_hostname',
                    self.instance_host)
            setattr(fake_server, 'OS-EXT-STS:vm_state', status)

            return fake_server

        failed_evacuation_instances = []
        fake_instance = self.fake_client.servers.create(
            id="1", host=self.instance_host, ha_enabled=True)

        _mock_get.side_effect = [
            get_fake_server(fake_instance, 'active'),
            get_fake_server(fake_instance, 'error'),
        ]
        task._evacuate_and_confirm(self.ctxt, fake_instance,
                                   self.instance_host,
                                   failed_evacuation_instances)
        self.assertIn(fake_instance.id, failed_evacuation_instances)
        expected_log = 'Failed to evacuate instance %s' % fake_instance.id
        _mock_log.warning.assert_called_once_with(expected_log)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_auto_recovery(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             True, "host_failure")

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host)

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(2)

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '1'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '2'", 0.8),
            mock.call("Instances to be evacuated are: '1,2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2'"),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '1,2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_auto_recovery_custom_ha_key(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             False, "host_failure")

        ha_enabled_key = 'Ensure-My-HA'

        self.override_config('ha_enabled_instance_metadata_key',
                             ha_enabled_key, 'host_failure')

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host,
                                        ha_enabled=True,
                                        ha_enabled_key=ha_enabled_key)

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(1)

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Instances to be evacuated are: '2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '2'"),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_reserved_host_recovery(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             True, "host_failure")
        self.override_config("add_reserved_host_to_aggregate",
                             True, "host_failure")

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host)
        reserved_host = fakes.create_fake_host(name="fake-reserved-host",
                                               reserved=True)
        self.fake_client.aggregates.create(id="1", name='fake_agg',
                                           hosts=[self.instance_host])

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(2)

        # execute EvacuateInstancesTask
        with mock.patch.object(manager, "update_host_method") as mock_save:
            self._evacuate_instances(
                instance_list, mock_enable_disable,
                reserved_host=reserved_host.name)
            self.assertEqual(1, mock_save.call_count)
            self.assertIn(reserved_host.name,
                          self.fake_client.aggregates.get('1').hosts)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '1'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '2'", 0.8),
            mock.call("Instances to be evacuated are: '1,2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2'"),
            mock.call("Enabling reserved host: 'fake-reserved-host'", 0.1),
            mock.call('Add host fake-reserved-host to aggregate fake_agg',
                      0.2),
            mock.call('Added host fake-reserved-host to aggregate fake_agg',
                      0.3),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '1,2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_multiple_aggregates(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             True, "host_failure")
        self.override_config("add_reserved_host_to_aggregate",
                             True, "host_failure")

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host)
        reserved_host = fakes.create_fake_host(name="fake-reserved-host",
                                               reserved=True)
        # Set multiple aggregates to the failure host
        self.fake_client.aggregates.create(id="1", name='fake_agg_1',
                                           hosts=[self.instance_host])
        self.fake_client.aggregates.create(id="2", name='fake_agg_2',
                                           hosts=[self.instance_host])

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(2)

        # execute EvacuateInstancesTask
        with mock.patch.object(manager, "update_host_method") as mock_save:
            self._evacuate_instances(
                instance_list, mock_enable_disable,
                reserved_host=reserved_host.name)
            self.assertEqual(1, mock_save.call_count)
            self.assertIn(reserved_host.name,
                          self.fake_client.aggregates.get('1').hosts)
            self.assertIn(reserved_host.name,
                          self.fake_client.aggregates.get('2').hosts)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '1'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '2'", 0.8),
            mock.call("Instances to be evacuated are: '1,2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2'"),
            mock.call("Enabling reserved host: 'fake-reserved-host'", 0.1),
            mock.call('Add host fake-reserved-host to aggregate fake_agg_1',
                      0.2),
            mock.call('Added host fake-reserved-host to aggregate fake_agg_1',
                      0.3),
            mock.call('Add host fake-reserved-host to aggregate fake_agg_2',
                      0.2),
            mock.call('Added host fake-reserved-host to aggregate fake_agg_2',
                      0.3),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '1,2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch.object(nova.API, 'add_host_to_aggregate')
    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.LOG')
    def test_host_failure_flow_ignores_conflict_error(
            self, mock_log, _mock_notify, _mock_novaclient, mock_add_host,
            mock_unlock, mock_lock, mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        mock_add_host.side_effect = exception.Conflict
        self.override_config("add_reserved_host_to_aggregate",
                             True, "host_failure")

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        reserved_host = fakes.create_fake_host(name="fake-reserved-host",
                                               reserved=True)
        self.fake_client.aggregates.create(id="1", name='fake_agg',
                                           hosts=[self.instance_host,
                                                  reserved_host.name])
        expected_msg_format = ("Host '%(reserved_host)s' already has been "
                               "added to aggregate '%(aggregate)s'.") % {
            'reserved_host': 'fake-reserved-host', 'aggregate': 'fake_agg'
        }

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(1)

        # execute EvacuateInstancesTask
        with mock.patch.object(manager, "update_host_method") as mock_save:
            self._evacuate_instances(
                instance_list, mock_enable_disable,
                reserved_host=reserved_host.name)
            self.assertEqual(1, mock_save.call_count)
            self.assertIn(reserved_host.name,
                          self.fake_client.aggregates.get('1').hosts)
            mock_log.info.assert_any_call(expected_msg_format)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 1"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Instances to be evacuated are: '1'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1'"),
            mock.call("Enabling reserved host: 'fake-reserved-host'", 0.1),
            mock.call('Add host fake-reserved-host to aggregate fake_agg',
                      0.2),
            mock.call("Host 'fake-reserved-host' already has been added to "
                      "aggregate 'fake_agg'.", 1.0),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Successfully evacuate instances '1' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @ddt.data('rescued', 'paused', 'shelved', 'suspended',
              'error', 'resized', 'active', 'resized', 'stopped')
    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_all_instances(
            self, vm_state, _mock_notify, _mock_novaclient, mock_unlock,
            mock_lock, mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        power_state = 4 if vm_state == 'resized' else None
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        vm_state=vm_state,
                                        power_state=power_state,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host,
                                        vm_state=vm_state,
                                        power_state=power_state,
                                        ha_enabled=True)

        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": instance_uuid_list,
        }

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2'"),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '1,2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_ignore_error_instances(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        self.override_config("ignore_instances_in_error_state",
                             True, "host_failure")
        self.override_config("evacuate_all_instances",
                             True, "host_failure")
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        vm_state='error',
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host,
                                        vm_state='active',
                                        ha_enabled=True)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(1)

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Ignoring recovery of HA_Enabled instance '1' as it is "
                      "in 'error' state.", 0.4),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '0'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '1'", 0.8),
            mock.call("Instances to be evacuated are: '2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '2'"),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_ignore_error_instances_raise_skip_host_recovery(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        self.override_config("ignore_instances_in_error_state",
                             True, "host_failure")
        self.override_config("evacuate_all_instances",
                             False, "host_failure")
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        vm_state='error',
                                        ha_enabled=True)

        # execute PrepareHAEnabledInstancesTask
        task = host_failure.PrepareHAEnabledInstancesTask(self.ctxt,
                                                          self.novaclient)
        self.assertRaises(exception.SkipHostRecoveryException, task.execute,
                          self.instance_host)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 1"
                      "", 0.3),
            mock.call("Ignoring recovery of HA_Enabled instance '1' as it is "
                      "in 'error' state.", 0.4),
            mock.call("Total HA Enabled instances count: '0'", 0.6),
            mock.call("Skipped host 'fake-host' recovery as no instances needs"
                      " to be evacuated", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_no_ha_enabled_instances(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host)
        self.fake_client.servers.create(id="2", host=self.instance_host)

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        task = host_failure.PrepareHAEnabledInstancesTask(self.ctxt,
                                                          self.novaclient)
        self.assertRaises(exception.SkipHostRecoveryException, task.execute,
                          self.instance_host)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '0'", 0.6),
            mock.call("Skipped host 'fake-host' recovery as no instances needs"
                      " to be evacuated", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_evacuation_failed(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        # overriding 'wait_period_after_power_off' to 2 seconds to reduce the
        # wait period, default is 180 seconds.
        self.override_config("wait_period_after_power_off", 2)
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        server = self.fake_client.servers.create(id="1", vm_state='active',
                                                 host=self.instance_host,
                                                 ha_enabled=True)

        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": instance_uuid_list,
        }

        def fake_get_server(context, host):
            # assume that while evacuating instance goes into error state
            fake_server = copy.deepcopy(server)
            setattr(fake_server, 'OS-EXT-STS:vm_state', "error")
            return fake_server

        with mock.patch.object(self.novaclient, "get_server", fake_get_server):
            # execute EvacuateInstancesTask
            self.assertRaises(
                exception.HostRecoveryFailureException,
                self._evacuate_instances, instance_list, mock_enable_disable)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1'"),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Failed to evacuate instances '1' from host 'fake-host'"
                      "", 0.7)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_no_instances_on_host(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             True, "host_failure")

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        task = host_failure.PrepareHAEnabledInstancesTask(self.ctxt,
                                                          self.novaclient)
        self.assertRaises(exception.SkipHostRecoveryException, task.execute,
                          self.instance_host)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 0"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '0'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '0'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '0'", 0.8),
            mock.call("Skipped host 'fake-host' recovery as no instances needs"
                      " to be evacuated", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_task_state_not_none(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        vm_state='active',
                                        task_state='fake_task_state',
                                        power_state=None,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host,
                                        vm_state='stopped',
                                        task_state='fake_task_state',
                                        power_state=None,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="3", host=self.instance_host,
                                        vm_state='error',
                                        task_state='fake_task_state',
                                        power_state=None,
                                        ha_enabled=True)
        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": instance_uuid_list,
        }

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        reset_calls = [('1', 'active'),
                       ('2', 'stopped'),
                       ('3', 'error'),
                       ('3', 'stopped')]
        stop_calls = ['2', '3']
        self.assertEqual(reset_calls,
                         self.fake_client.servers.reset_state_calls)
        self.assertEqual(stop_calls,
                         self.fake_client.servers.stop_calls)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2,3'"),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Evacuation of instance started: '3'", 0.5),
            mock.call("Successfully evacuate instances '1,2,3' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_host_failure_flow_for_RH_recovery(
            self, _mock_notify, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client
        self.override_config("evacuate_all_instances",
                             True, "host_failure")
        self.override_config("add_reserved_host_to_aggregate",
                             True, "host_failure")

        # create test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host=self.instance_host)
        reserved_host = fakes.create_fake_host(name="fake-reserved-host",
                                               reserved=True)
        self.fake_client.aggregates.create(id="1", name='fake_agg',
                                           hosts=[self.instance_host])

        # execute DisableComputeServiceTask
        self._test_disable_compute_service(mock_enable_disable)

        # execute PrepareHAEnabledInstancesTask
        instance_list = self._test_instance_list(2)

        # execute EvacuateInstancesTask
        with mock.patch.object(manager, "update_host_method") as mock_save:
            self._evacuate_instances(
                instance_list, mock_enable_disable,
                reserved_host=reserved_host.name)
            self.assertEqual(1, mock_save.call_count)
            self.assertIn(reserved_host.name,
                          self.fake_client.aggregates.get('1').hosts)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call('Preparing instances for evacuation'),
            mock.call("Total instances running on failed host 'fake-host' is 2"
                      "", 0.3),
            mock.call("Total HA Enabled instances count: '1'", 0.6),
            mock.call("Total Non-HA Enabled instances count: '1'", 0.7),
            mock.call("All instances (HA Enabled/Non-HA Enabled) should be "
                      "considered for evacuation. Total count is: '2'", 0.8),
            mock.call("Instances to be evacuated are: '1,2'", 1.0),
            mock.call("Start evacuation of instances from failed host "
                      "'fake-host', instance uuids are: '1,2'"),
            mock.call("Enabling reserved host: 'fake-reserved-host'", 0.1),
            mock.call('Add host fake-reserved-host to aggregate fake_agg',
                      0.2),
            mock.call('Added host fake-reserved-host to aggregate fake_agg',
                      0.3),
            mock.call("Evacuation of instance started: '1'", 0.5),
            mock.call("Evacuation of instance started: '2'", 0.5),
            mock.call("Successfully evacuate instances '1,2' from host "
                      "'fake-host'", 0.7),
            mock.call('Evacuation process completed!', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    def test_host_failure_flow_for_stopped_instances(
            self, _mock_novaclient, mock_unlock, mock_lock,
            mock_enable_disable):
        _mock_novaclient.return_value = self.fake_client

        # create ha_enabled test data
        self.fake_client.servers.create(id="1", host=self.instance_host,
                                        vm_state='stopped',
                                        task_state=None,
                                        power_state=None,
                                        ha_enabled=True)
        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": instance_uuid_list,
        }

        # execute EvacuateInstancesTask
        self._evacuate_instances(instance_list, mock_enable_disable)

        # If vm_state=stopped and task_state=None, reset_state and stop API
        # will not be called.
        self.assertEqual(0, len(self.fake_client.servers.reset_state_calls))
        self.assertEqual(0, len(self.fake_client.servers.stop_calls))



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from oslo_config import cfg

service_opts = [
    cfg.HostAddressOpt('host',
               default=socket.gethostname(),
               help='''
Hostname, FQDN or IP address of this host. Must be valid within AMQP key.

Possible values:

* String with hostname, FQDN or IP address. Default is hostname of this host.
'''),
    cfg.StrOpt('engine_manager',
               default='masakari.engine.manager.MasakariManager',
               help='Full class name for the Manager for masakari engine'),
    cfg.IntOpt('report_interval',
               default=10,
               help='Seconds between nodes reporting state to datastore'),
    cfg.BoolOpt('periodic_enable',
                default=True,
                help='Enable periodic tasks'),
    cfg.IntOpt('periodic_interval_max',
               default=300,
               help='Max interval time between periodic tasks execution in '
                    'seconds.'),
    cfg.IntOpt('periodic_fuzzy_delay',
               default=60,
               help='Range of seconds to randomly delay when starting the'
                    ' periodic task scheduler to reduce stampeding.'
                    ' (Disable by setting to 0)'),
    cfg.BoolOpt('use_ssl',
                default=False,
                help='Use APIs with SSL enabled'),
    cfg.HostAddressOpt('masakari_api_listen',
               default="0.0.0.0",
               help='The IP address on which the Masakari API will listen.'),
    cfg.IntOpt('masakari_api_listen_port',
               default=15868,
               min=1,
               max=65535,
               help='The port on which the Masakari API will listen.'),
    cfg.IntOpt('masakari_api_workers',
               help='Number of workers for Masakari API service. The default '
                    'will be the number of CPUs available.'),
    cfg.IntOpt('service_down_time',
               default=60,
               help='Maximum time since last check-in for up service'),
    ]


def register_opts(conf):
    conf.register_opts(service_opts)


def list_opts():
    return {'DEFAULT': service_opts}



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from migrate.changeset import UniqueConstraint
from sqlalchemy import Column, MetaData, Table
from sqlalchemy import Integer, DateTime, String, Enum, Text


def define_notifications_table(meta):

    notifications = Table('notifications',
                          meta,
                          Column('created_at', DateTime),
                          Column('updated_at', DateTime),
                          Column('deleted_at', DateTime),
                          Column('deleted', Integer),
                          Column('id', Integer, primary_key=True,
                                 nullable=False),
                          Column('notification_uuid', String(36),
                                 nullable=False),
                          Column('generated_time', DateTime, nullable=False),
                          Column('source_host_uuid', String(36), nullable=False
                                 ),
                          Column('type', String(length=36), nullable=False),
                          Column('payload', Text),
                          Column('status',
                                 Enum('new', 'running', 'error', 'failed',
                                      'ignored', 'finished',
                                      name='notification_status'),
                                 nullable=False),
                          UniqueConstraint('notification_uuid',
                                           name='uniq_notifications0uuid'),
                          mysql_engine='InnoDB',
                          mysql_charset='utf8',
                          extend_existing=True)

    return notifications


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    table = define_notifications_table(meta)
    table.create()



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import iso8601
from oslo_utils import timeutils
from oslo_utils import uuidutils

from masakari import objects
from masakari.objects import segment as segment_obj
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)


def _make_segment_obj(segment_dict):
    return segment_obj.FailoverSegment(**segment_dict)


FAILOVER_SEGMENT = {"name": "segment1", "id": "1",
                    "service_type": "COMPUTE", "recovery_method": "auto",
                    "uuid": uuidsentinel.fake_segment1,
                    "description": "failover_segment for compute"}


FAILOVER_SEGMENT = _make_segment_obj(FAILOVER_SEGMENT)


class FakeNovaClient(object):
    class Server(object):
        def __init__(self, id=None, uuid=None, host=None, vm_state=None,
                     task_state=None, power_state=1, ha_enabled=None,
                     ha_enabled_key='HA_Enabled', locked=False):
            self.id = id
            self.uuid = uuid or uuidutils.generate_uuid()
            self.host = host
            setattr(self, 'OS-EXT-SRV-ATTR:hypervisor_hostname', host)
            setattr(self, 'OS-EXT-STS:vm_state', vm_state)
            setattr(self, 'OS-EXT-STS:task_state', task_state)
            setattr(self, 'OS-EXT-STS:power_state', power_state)
            self.metadata = {ha_enabled_key: ha_enabled}
            self.locked = locked

    class ServerManager(object):
        def __init__(self):
            self._servers = []
            self.reset_state_calls = []
            self.stop_calls = []

        def create(self, id, uuid=None, host=None, vm_state='active',
                   task_state=None, power_state=1, ha_enabled=False,
                   ha_enabled_key='HA_Enabled'):
            server = FakeNovaClient.Server(id=id, uuid=uuid, host=host,
                                           vm_state=vm_state,
                                           task_state=task_state,
                                           power_state=power_state,
                                           ha_enabled=ha_enabled,
                                           ha_enabled_key=ha_enabled_key)
            self._servers.append(server)
            return server

        def get(self, id):
            for s in self._servers:
                if s.id == id:
                    return s
            return None

        def list(self, detailed=True, search_opts=None):
            matching = list(self._servers)
            if search_opts:
                for opt, val in search_opts.items():
                    if 'all_tenants' in search_opts:
                        continue
                    matching = [m for m in matching
                                if getattr(m, opt, None) == val]
            return matching

        def reset_state(self, uuid, status):
            current_status = getattr(self.get(uuid), "OS-EXT-STS:vm_state")
            self.reset_state_calls.append((uuid, current_status))
            server = self.get(uuid)
            setattr(server, 'OS-EXT-STS:vm_state', status)

        def evacuate(self, uuid, host=None):
            if not host:
                host = 'fake-host-1'
            server = self.get(uuid)
            setattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname', host)
            # pretending that instance is evacuated successfully on given host
            if getattr(server, "OS-EXT-STS:vm_state") in ['active', 'error']:
                setattr(server, 'OS-EXT-STS:vm_state', 'active')
            else:
                setattr(server, 'OS-EXT-STS:vm_state', 'stopped')

        def stop(self, id):
            self.stop_calls.append(id)
            server = self.get(id)
            setattr(server, 'OS-EXT-STS:vm_state', 'stopped')

        def start(self, id):
            server = self.get(id)
            setattr(server, 'OS-EXT-STS:vm_state', 'active')

    class Aggregate(object):
        def __init__(self, id=None, uuid=None, name=None, hosts=None):
            self.id = id
            self.uuid = uuid or uuidutils.generate_uuid()
            self.name = name
            self.hosts = hosts

    class AggregatesManager(object):
        def __init__(self):
            self.aggregates = []

        def create(self, id, uuid=None, name=None, hosts=None):
            aggregate = FakeNovaClient.Aggregate(id=id, uuid=uuid, name=name,
                                                 hosts=hosts)
            self.aggregates.append(aggregate)
            return aggregate

        def list(self):
            return self.aggregates

        def add_host(self, aggregate_id, host_name):
            aggregate = self.get(aggregate_id)
            if host_name not in aggregate.hosts:
                aggregate.hosts.append(host_name)

        def get(self, aggregate_id):
            for aggregate in self.aggregates:
                if aggregate.id == aggregate_id:
                    return aggregate

    class Service(object):
        def __init__(self, id=None, host=None, binary=None, status='enabled'):
            self.id = id
            self.host = host
            self.binary = binary
            self.status = status

    class Services(object):
        def __init__(self):
            self._services = []

        def create(self, id, host=None, binary=None,
                   status=None):
            self._services.append(FakeNovaClient.Service(id=id, host=host,
                                                         binary=binary,
                                                         status=status))

        def disable(self, service_id):
            for _service in self._services:
                if _service.id == service_id:
                    service = _service
            service.status = 'disabled'

        def list(self, host=None, binary=None):
            services = []
            for service in self._services:
                if host == service.host and binary == service.binary:
                    services.append(service)
            return services

    def __init__(self):
        self.servers = FakeNovaClient.ServerManager()
        self.services = FakeNovaClient.Services()
        self.aggregates = FakeNovaClient.AggregatesManager()


def create_fake_notification(type="VM", id=1, payload=None,
                             source_host_uuid=uuidsentinel.fake_host,
                             generated_time=NOW, status="new",
                             notification_uuid=uuidsentinel.fake_notification):
    return objects.Notification(
        type=type, id=id, payload=payload, source_host_uuid=source_host_uuid,
        generated_time=generated_time, status=status,
        notification_uuid=notification_uuid)


def create_fake_host(**updates):
    host = {
        'name': 'fake_host', 'id': 1, 'reserved': False,
        'on_maintenance': False, 'type': 'SSH',
        'control_attributes': 'fake', 'uuid': uuidsentinel.fake_host,
        'failover_segment': FAILOVER_SEGMENT,
        'created_at': datetime.datetime(
            2019, 8, 8, 0, 0, 0, tzinfo=iso8601.UTC),
        'updated_at': None, 'deleted_at': None, 'deleted': False
    }
    if updates:
        host.update(updates)
    return objects.Host(**host)


def create_fake_failover_segment(name='fake_segment', id=1, description=None,
                                 service_type='COMPUTE',
                                 recovery_method="auto",
                                 uuid=uuidsentinel.fake_segment,
                                 enabled=True):
    return objects.FailoverSegment(
        name=name, id=id, description=description, service_type=service_type,
        recovery_method=recovery_method, uuid=uuid, enabled=enabled)


def create_fake_notification_progress_details(
        name, uuid, progress, state, progress_details):

    return objects.NotificationProgressDetails(
        name=name, uuid=uuid, progress=progress, state=state,
        progress_details=progress_details)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Common parameter types for validating request Body.

"""

import re
import unicodedata

from masakari.i18n import _


class ValidationRegex(object):
    def __init__(self, regex, reason):
        self.regex = regex
        self.reason = reason


def _is_printable(char):
    """determine if a unicode code point is printable.

    This checks if the character is either "other" (mostly control
    codes), or a non-horizontal space. All characters that don't match
    those criteria are considered printable; that is: letters;
    combining marks; numbers; punctuation; symbols; (horizontal) space
    separators.
    """
    category = unicodedata.category(char)
    return (not category.startswith("C") and
            (not category.startswith("Z") or category == "Zs"))


def _get_all_chars():
    for i in range(0xFFFF):
        yield chr(i)


# build a regex that matches all printable characters. This allows
# spaces in the middle of the name. Also note that the regexp below
# deliberately allows the empty string. This is so only the constraint
# which enforces a minimum length for the name is triggered when an
# empty string is tested. Otherwise it is not deterministic which
# constraint fails and this causes issues for some unittests when
# PYTHONHASHSEED is set randomly.

def _build_regex_range(ws=True, invert=False, exclude=None):
    """Build a range regex for a set of characters in utf8.

    This builds a valid range regex for characters in utf8 by
    iterating the entire space and building up a set of x-y ranges for
    all the characters we find which are valid.

    :param ws: should we include whitespace in this range.
    :param exclude: any characters we want to exclude
    :param invert: invert the logic

    The inversion is useful when we want to generate a set of ranges
    which is everything that's not a certain class. For instance,
    produce all all the non printable characters as a set of ranges.
    """
    if exclude is None:
        exclude = []
    regex = ""
    # are we currently in a range
    in_range = False
    # last character we found, for closing ranges
    last = None
    # last character we added to the regex, this lets us know that we
    # already have B in the range, which means we don't need to close
    # it out with B-B. While the later seems to work, it's kind of bad form.
    last_added = None

    def valid_char(char):
        if char in exclude:
            result = False
        elif ws:
            result = _is_printable(char)
        else:
            # Zs is the unicode class for space characters, of which
            # there are about 10 in this range.
            result = (_is_printable(char) and
                      unicodedata.category(char) != "Zs")
        if invert is True:
            return not result
        return result

    # iterate through the entire character range. in_
    for c in _get_all_chars():
        if valid_char(c):
            if not in_range:
                regex += re.escape(c)
                last_added = c
            in_range = True
        else:
            if in_range and last != last_added:
                regex += "-" + re.escape(last)
            in_range = False
        last = c
    else:
        if in_range:
            regex += "-" + re.escape(c)
    return regex


valid_description_regex_base = '^[\n%s]*$'

valid_name_regex_base = '^(?![%s])[%s]*(?<![%s])$'

valid_description_regex = valid_description_regex_base % (
    _build_regex_range())

valid_name_regex = ValidationRegex(
    valid_name_regex_base % (
        _build_regex_range(ws=False, invert=True),
        _build_regex_range(),
        _build_regex_range(ws=False, invert=True)),
    _("printable characters. Can not start or end with whitespace."))

name = {
    'type': 'string', 'minLength': 1, 'maxLength': 255,
    'format': 'name'
}


description = {
    'type': ['string', 'null'], 'minLength': 0,
    'pattern': valid_description_regex,
}


type = {
    'type': ['string', 'null'], 'minLength': 0, 'maxLength': 255,
    'pattern': '^[a-zA-Z0-9]*$',
}


boolean = {
    'type': ['boolean', 'string'],
    'enum': [True, 'True', 'TRUE', 'true', '1', 'ON', 'On', 'on',
             'YES', 'Yes', 'yes',
             False, 'False', 'FALSE', 'false', '0', 'OFF', 'Off', 'off',
             'NO', 'No', 'no'],
}


hostname = {
    'type': 'string', 'minLength': 1, 'maxLength': 255,
    'pattern': '^[a-zA-Z0-9-._]*$',
}


payload = {
    'type': 'object',
    'patternProperties': {
        '^[a-zA-Z0-9-_:. ]{1,255}$': {
            'type': 'string', 'maxLength': 255
        }
    },
    'additionalProperties': False
}



# Copyright 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Handles all requests to Nova.
"""

import functools
import sys

from keystoneauth1 import exceptions as keystone_exception
import keystoneauth1.loading
import keystoneauth1.loading.session
from novaclient import api_versions
from novaclient import client as nova_client
from novaclient import exceptions as nova_exception
from oslo_log import log as logging
from oslo_utils import encodeutils
from requests import exceptions as request_exceptions

from masakari import conf
from masakari import context as ctx
from masakari import exception
from masakari import utils

CONF = conf.CONF
CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')

LOG = logging.getLogger(__name__)

NOVA_API_VERSION = "2.53"

nova_extensions = [ext for ext in
                   nova_client.discover_extensions(NOVA_API_VERSION)
                   if ext.name in ("list_extensions",)]


def _reraise(desired_exc):
    utils.reraise(type(desired_exc), desired_exc, sys.exc_info()[2])


def translate_nova_exception(method):
    """Transforms a cinder exception but keeps its traceback intact."""
    @functools.wraps(method)
    def wrapper(self, ctx, *args, **kwargs):
        try:
            res = method(self, ctx, *args, **kwargs)
        except (request_exceptions.Timeout,
                nova_exception.CommandError,
                keystone_exception.ConnectionError) as exc:
            err_msg = encodeutils.exception_to_unicode(exc)
            _reraise(exception.MasakariException(reason=err_msg))
        except (keystone_exception.BadRequest,
                nova_exception.BadRequest) as exc:
            err_msg = encodeutils.exception_to_unicode(exc)
            _reraise(exception.InvalidInput(reason=err_msg))
        except (keystone_exception.Forbidden,
                nova_exception.Forbidden) as exc:
            err_msg = encodeutils.exception_to_unicode(exc)
            _reraise(exception.Forbidden(err_msg))
        except (nova_exception.NotFound) as exc:
            err_msg = encodeutils.exception_to_unicode(exc)
            _reraise(exception.NotFound(reason=err_msg))
        except nova_exception.Conflict as exc:
            err_msg = encodeutils.exception_to_unicode(exc)
            _reraise(exception.Conflict(reason=err_msg))
        return res
    return wrapper


def novaclient(context, timeout=None):
    """Returns a Nova client

    @param timeout: Number of seconds to wait for an answer before raising a
        Timeout exception (None to disable)
    """
    nova_catalog_info = CONF.nova_catalog_admin_info
    service_type, service_name, endpoint_type = nova_catalog_info.split(':')

    context = ctx.RequestContext(
        CONF.os_privileged_user_name, None,
        auth_token=CONF.os_privileged_user_password,
        project_name=CONF.os_privileged_user_tenant,
        service_catalog=context.service_catalog,
        global_request_id=context.global_id)

    # User needs to authenticate to Keystone before querying Nova, so we set
    # auth_url to the identity service endpoint
    url = CONF.os_privileged_user_auth_url

    LOG.debug('Creating a Nova client using "%s" user',
              CONF.os_privileged_user_name)

    # Now that we have the correct auth_url, username, password and
    # project_name, let's build a Keystone session.
    loader = keystoneauth1.loading.get_plugin_loader(
        CONF.keystone_authtoken.auth_type)
    auth = loader.load_from_options(
        auth_url=url,
        username=context.user_id,
        password=context.auth_token,
        project_name=context.project_name,
        user_domain_name=CONF.os_user_domain_name,
        project_domain_name=CONF.os_project_domain_name,
        system_scope=CONF.os_system_scope)
    session_loader = keystoneauth1.loading.session.Session()
    keystone_session = session_loader.load_from_options(
        auth=auth, cacert=CONF.nova_ca_certificates_file,
        insecure=CONF.nova_api_insecure)

    client_obj = nova_client.Client(
        api_versions.APIVersion(NOVA_API_VERSION),
        session=keystone_session,
        insecure=CONF.nova_api_insecure,
        timeout=timeout,
        global_request_id=context.global_id,
        region_name=CONF.os_region_name,
        endpoint_type=endpoint_type,
        service_type=service_type,
        service_name=service_name,
        cacert=CONF.nova_ca_certificates_file,
        extensions=nova_extensions)

    return client_obj


class API(object):
    """API for interacting with novaclient."""

    @translate_nova_exception
    def get_servers(self, context, host):
        """Get a list of servers running on a specified host."""
        opts = {
            'host': host,
            'all_tenants': True
        }
        nova = novaclient(context)
        LOG.info('Fetch Server list on %s', host)
        return nova.servers.list(detailed=True, search_opts=opts)

    @translate_nova_exception
    def enable_disable_service(self, context, host_name, enable=False,
                               reason=None):
        """Enable or disable the service specified by nova service id."""
        nova = novaclient(context)
        service = nova.services.list(host=host_name, binary='nova-compute')[0]

        if not enable:
            LOG.info('Disable nova-compute on %s', host_name)
            if reason:
                nova.services.disable_log_reason(service.id, reason)
            else:
                nova.services.disable(service.id)
        else:
            LOG.info('Enable nova-compute on %s', host_name)
            nova.services.enable(service.id)

    @translate_nova_exception
    def is_service_down(self, context, host_name, binary):
        """Check whether service is up or down on given host."""
        nova = novaclient(context)
        service = nova.services.list(host=host_name, binary=binary)[0]
        return service.status == 'disabled'

    @translate_nova_exception
    def evacuate_instance(self, context, uuid, target=None):
        """Evacuate an instance from failed host to specified host."""
        msg = ('Call evacuate command for instance %(uuid)s on host '
               '%(target)s')
        LOG.info(msg, {'uuid': uuid, 'target': target})
        nova = novaclient(context)
        nova.servers.evacuate(uuid, host=target)

    @translate_nova_exception
    def reset_instance_state(self, context, uuid, status='error'):
        """Reset the state of an instance to active or error."""
        msg = ('Call reset state command on instance %(uuid)s to '
               'status: %(status)s.')
        LOG.info(msg, {'uuid': uuid, 'status': status})
        nova = novaclient(context)
        nova.servers.reset_state(uuid, status)

    @translate_nova_exception
    def get_server(self, context, uuid):
        """Get a server."""
        nova = novaclient(context)
        msg = ('Call get server command for instance %(uuid)s')
        LOG.info(msg, {'uuid': uuid})
        return nova.servers.get(uuid)

    @translate_nova_exception
    def stop_server(self, context, uuid):
        """Stop a server."""
        nova = novaclient(context)
        msg = ('Call stop server command for instance %(uuid)s')
        LOG.info(msg, {'uuid': uuid})
        return nova.servers.stop(uuid)

    @translate_nova_exception
    def start_server(self, context, uuid):
        """Start a server."""
        nova = novaclient(context)
        msg = ('Call start server command for instance %(uuid)s')
        LOG.info(msg, {'uuid': uuid})
        return nova.servers.start(uuid)

    @translate_nova_exception
    def get_aggregate_list(self, context):
        """Get all aggregate list."""
        nova = novaclient(context)
        LOG.info('Call aggregate-list command to get list of all aggregates.')
        return nova.aggregates.list()

    @translate_nova_exception
    def add_host_to_aggregate(self, context, host, aggregate):
        """Add host to given aggregate."""
        nova = novaclient(context)
        msg = ("Call add_host command to add host '%(host_name)s' to "
               "aggregate '%(aggregate_name)s'.")
        LOG.info(msg, {'host_name': host, 'aggregate_name': aggregate.name})
        return nova.aggregates.add_host(aggregate.id, host)

    @translate_nova_exception
    def lock_server(self, context, uuid):
        """Lock a server."""
        nova = novaclient(context)
        msg = ('Call lock server command for instance %(uuid)s')
        LOG.info(msg, {'uuid': uuid})
        return nova.servers.lock(uuid)

    @translate_nova_exception
    def unlock_server(self, context, uuid):
        """Unlock a server."""
        nova = novaclient(context)
        msg = ('Call unlock server command for instance %(uuid)s')
        LOG.info(msg, {'uuid': uuid})
        return nova.servers.unlock(uuid)

    @translate_nova_exception
    def find_compute_service(self, context, compute_name):
        """Find compute service with case sensitive hostname."""
        nova = novaclient(context)
        msg = ("Call compute service find command to get list of matching "
               "hypervisor name '%(compute_name)s'")
        LOG.info(msg, {'compute_name': compute_name})

        computes = \
            nova.services.list(binary='nova-compute', host=compute_name)
        if len(computes) == 0:
            raise exception.ComputeNotFoundByName(
                compute_name=compute_name)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from masakari.objects import fields
from masakari.tests.functional import base


@ddt.ddt
class TestHosts(base.BaseFunctionalTest):

    def setUp(self):
        super(TestHosts, self).setUp()

        if not self.hypervisors:
            self.skipTest("Skipped as there are no hypervisors "
                          "configured in nova")

        # Create segment
        self.segment = self.admin_conn.ha.create_segment(
            name=self.getUniqueString(),
            recovery_method=fields.FailoverSegmentRecoveryMethod.AUTO,
            service_type='COMPUTE')

        # Delete segment which deletes host/s associated with it
        self.addCleanup(self.admin_conn.ha.delete_segment, self.segment.uuid)

    def test_create_get(self):
        # This test is for testing hosts create/get
        # Create valid host
        host_name = self.hypervisors[0]['name']
        host_data = {'name': host_name,
                     'type': 'COMPUTE',
                     'on_maintenance': False,
                     'reserved': False,
                     'control_attributes': 'SSH'}

        host = self.admin_conn.ha.create_host(self.segment.uuid, **host_data)

        self.assertDictContainsSubset(host_data, host)

        result = self.admin_conn.ha.get_host(host.uuid, self.segment.uuid)

        self.assertEqual('COMPUTE', result.type)
        self.assertEqual(False, result.on_maintenance)
        self.assertEqual(False, result.reserved)
        self.assertEqual('SSH', result.control_attributes)

    def test_list(self):
        # This test is for testing host/s creation and listing the same.

        expected_hosts = []
        for host in self.hypervisors:
            host_data = {
                'name': host.name,
                'type': 'COMPUTE',
                'on_maintenance': False,
                'reserved': False,
                'control_attributes': 'SSH',
            }
            self.admin_conn.ha.create_host(
                segment_id=self.segment.uuid,
                **host_data)

            # NOTE(yoctozepto): 'failover_segment_id' is added in the API
            # response. We can verify it here.
            host_data['failover_segment_id'] = self.segment.uuid

            expected_hosts.append(host_data)

        hosts = self.admin_conn.ha.hosts(self.segment.uuid)
        # NOTE(yoctozepto): We are saving the generator values to a list to
        # compare the length and then iterate over the elements for comparison.
        hosts = list(hosts)
        self.assertEqual(len(expected_hosts), len(hosts))
        for expected_host in expected_hosts:
            found = False
            for host in hosts:
                found = found or (dict(host, **expected_host) == dict(host))
            self.assertEqual(True, found,
                'Host not found: {expected_host}'.format(
                    expected_host=expected_host))

    @ddt.data(
        {'on_maintenance': False, 'host_type': 'COMPUTE', 'reserved': False,
         'control_attributes': 'SSH'},
        {'on_maintenance': True, 'host_type': 'CONTROLLER', 'reserved': True,
         'control_attributes': 'TCP'}
    )
    @ddt.unpack
    def test_create_list_with_filter(self, on_maintenance,
                                     host_type, reserved, control_attributes):
        # This test is for testing host/s creation and listing
        # the same based on filters.
        if len(self.hypervisors) == 1:
            self.skipTest("Skipped as there is only one hypervisor "
                          "configured in nova")

        host_data_1 = {'name': self.hypervisors[0].name,
                       'type': 'COMPUTE',
                       'on_maintenance': False,
                       'reserved': False,
                       'control_attributes': 'SSH'}

        host_data_2 = {'name': self.hypervisors[1].name,
                       'type': 'CONTROLLER',
                       'on_maintenance': True,
                       'reserved': True,
                       'control_attributes': 'TCP'}

        self.admin_conn.ha.create_host(self.segment.uuid, **host_data_1)
        self.admin_conn.ha.create_host(self.segment.uuid, **host_data_2)

        expected_host_data = {'on_maintenance': on_maintenance,
                              'type': host_type,
                              'reserved': reserved,
                              'control_attributes': control_attributes
                              }

        # Returns list of hosts based on filters
        for host in self.admin_conn.ha.hosts(self.segment.uuid,
                                       on_maintenance=on_maintenance,
                                       type=host_type,
                                       reserved=reserved):

            self.assertDictContainsSubset(expected_host_data, host)

    def test_update_get_delete(self):
        # This test is for updating created host and deletion of same
        host_name = self.hypervisors[0]['name']

        host = self.admin_conn.ha.create_host(segment_id=self.segment.uuid,
                                        name=host_name,
                                        on_maintenance='False',
                                        reserved='False',
                                        type='COMPUTE',
                                        control_attributes='SSH')

        self.admin_conn.ha.update_host(host['uuid'],
                                 segment_id=self.segment.uuid,
                                 on_maintenance='True',
                                 control_attributes='TCP',
                                 reserved='True')

        result = self.admin_conn.ha.get_host(host.uuid,
                                       self.segment.uuid)
        # Confirm host update
        self.assertEqual(True, result.on_maintenance)
        self.assertEqual(True, result.reserved)
        self.assertEqual('TCP', result.control_attributes)

    def test_update_host_name(self):
        # This test is for updating host name
        if len(self.hypervisors) == 1:
            self.skipTest("Skipped as there is only one hypervisor "
                          "configured in nova")

        host = self.admin_conn.ha.create_host(segment_id=self.segment.uuid,
                            name=self.hypervisors[0]['name'],
                            type='COMPUTE',
                            control_attributes='SSH')

        # Update host name
        updated_host = self.admin_conn.ha.update_host(host['uuid'],
                segment_id=self.segment.uuid,
                name=self.hypervisors[1]['name'])

        result = self.admin_conn.ha.get_host(host.uuid,
                                       self.segment.uuid)
        self.assertEqual(result.name, updated_host.name)



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Implementation of SQLAlchemy backend."""

import datetime
import sys

from oslo_db import api as oslo_db_api
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_log import log as logging
from oslo_utils import timeutils
from sqlalchemy import or_, and_
from sqlalchemy.ext.compiler import compiles
from sqlalchemy import MetaData
from sqlalchemy.orm import joinedload
from sqlalchemy import sql
import sqlalchemy.sql as sa_sql
from sqlalchemy.sql import func

import masakari.conf
from masakari.db.sqlalchemy import models
from masakari import exception
from masakari.i18n import _

LOG = logging.getLogger(__name__)

CONF = masakari.conf.CONF

main_context_manager = enginefacade.transaction_context()


def _get_db_conf(conf_group, connection=None):
    kw = dict(conf_group.items())
    if connection is not None:
        kw['connection'] = connection
    return kw


def _context_manager_from_context(context):
    if context:
        try:
            return context.db_connection
        except AttributeError:
            pass


def get_backend():
    """The backend is this module itself."""
    return sys.modules[__name__]


def configure(conf):
    main_context_manager.configure(**_get_db_conf(conf.database))


def get_engine(use_slave=False, context=None):
    """Get a database engine object.

    :param use_slave: Whether to use the slave connection
    :param context: The request context that can contain a context manager
    """
    ctxt_mgr = _context_manager_from_context(context) or main_context_manager
    return ctxt_mgr.get_legacy_facade().get_engine(use_slave=use_slave)


def create_context_manager(connection=None):
    """Create a database context manager object.

    : param connection: The database connection string
    """
    ctxt_mgr = enginefacade.transaction_context()
    ctxt_mgr.configure(**_get_db_conf(CONF.database, connection=connection))
    return ctxt_mgr


def model_query(context, model, args=None, read_deleted=None):
    """Query helper that accounts for context's `read_deleted` field.
    :param context:     MasakariContext of the query.
    :param model:       Model to query. Must be a subclass of ModelBase.
    :param args:        Arguments to query. If None - model is used.
    :param read_deleted: If not None, overrides context's read_deleted field.
                        Permitted values are 'no', which does not return
                        deleted values; 'only', which only returns deleted
                        values; and 'yes', which does not filter deleted
                        values.
    """

    if read_deleted is None:
        read_deleted = context.read_deleted

    query_kwargs = {}
    if 'no' == read_deleted:
        query_kwargs['deleted'] = False
    elif 'only' == read_deleted:
        query_kwargs['deleted'] = True
    elif 'yes' == read_deleted:
        pass
    else:
        raise ValueError(_("Unrecognized read_deleted value '%s'")
                         % read_deleted)

    query = sqlalchemyutils.model_query(
        model, context.session, args, **query_kwargs)

    return query


def _process_sort_params(sort_keys, sort_dirs,
                         default_keys=['created_at', 'id'],
                         default_dir='desc'):
    """Process the sort parameters to include default keys.

    Creates a list of sort keys and a list of sort directions. Adds the default
    keys to the end of the list if they are not already included.

    When adding the default keys to the sort keys list, the associated
    direction is:
    1) The first element in the 'sort_dirs' list (if specified), else
    2) 'default_dir' value (Note that 'asc' is the default value since this is
    the default in sqlalchemy.utils.paginate_query)

    :param sort_keys: List of sort keys to include in the processed list
    :param sort_dirs: List of sort directions to include in the processed list
    :param default_keys: List of sort keys that need to be included in the
                         processed list, they are added at the end of the list
                         if not already specified.
    :param default_dir: Sort direction associated with each of the default
                        keys that are not supplied, used when they are added
                        to the processed list
    :returns: list of sort keys, list of sort directions
    :raise exception.InvalidInput: If more sort directions than sort keys
                                   are specified or if an invalid sort
                                   direction is specified
    """
    # Determine direction to use for when adding default keys
    default_dir_value = default_dir
    if sort_dirs and len(sort_dirs) != 0:
        default_dir_value = sort_dirs[0]

    # Create list of keys (do not modify the input list)
    result_keys = []
    if sort_keys:
        result_keys = list(sort_keys)

    # If a list of directions is not provided, use the default sort direction
    # for all provided keys
    if sort_dirs:
        result_dirs = []
        # Verify sort direction
        for sort_dir in sort_dirs:
            if sort_dir not in ('asc', 'desc'):
                msg = _("Unknown sort direction, must be 'asc' or 'desc'")
                raise exception.InvalidInput(reason=msg)
            result_dirs.append(sort_dir)
    else:
        result_dirs = [default_dir_value for _sort_key in result_keys]

    # Ensure that the key and direction length match
    while len(result_dirs) < len(result_keys):
        result_dirs.append(default_dir_value)
    # Unless more direction are specified, which is an error
    if len(result_dirs) > len(result_keys):
        msg = _("Sort direction size exceeds sort key size")
        raise exception.InvalidInput(reason=msg)

    # Ensure defaults are included
    for key in default_keys:
        if key not in result_keys:
            result_keys.append(key)
            result_dirs.append(default_dir_value)

    return result_keys, result_dirs


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def failover_segment_get_all_by_filters(
        context, filters=None, sort_keys=None,
        sort_dirs=None, limit=None, marker=None):

    # NOTE(Dinesh_Bhor): If the limit is 0 there is no point in even going
    # to the database since nothing is going to be returned anyway.
    if limit == 0:
        return []

    sort_keys, sort_dirs = _process_sort_params(sort_keys,
                                                sort_dirs)
    filters = filters or {}
    query = model_query(context, models.FailoverSegment)

    if 'recovery_method' in filters:
        query = query.filter(models.FailoverSegment.recovery_method == filters[
            'recovery_method'])
    if 'service_type' in filters:
        query = query.filter(models.FailoverSegment.service_type == filters[
            'service_type'])
    if 'enabled' in filters:
        query = query.filter(models.FailoverSegment.enabled == filters[
            'enabled'])

    marker_row = None
    if marker is not None:
        marker_row = model_query(context,
                                 models.FailoverSegment
                                 ).filter_by(id=marker).first()

        if not marker_row:
            raise exception.MarkerNotFound(marker=marker)

    try:
        query = sqlalchemyutils.paginate_query(query, models.FailoverSegment,
                                               limit, sort_keys,
                                               marker=marker_row,
                                               sort_dirs=sort_dirs)
    except db_exc.InvalidSortKey as e:
        raise exception.InvalidSortKey(e)

    return query.all()


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def failover_segment_get_by_id(context, segment_id):
    query = model_query(context,
                        models.FailoverSegment).filter_by(id=segment_id)

    result = query.first()
    if not result:
        raise exception.FailoverSegmentNotFound(id=segment_id)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def failover_segment_get_by_uuid(context, segment_uuid):
    return _failover_segment_get_by_uuid(context, segment_uuid)


def _failover_segment_get_by_uuid(context, segment_uuid):
    query = model_query(context,
                        models.FailoverSegment).filter_by(uuid=segment_uuid)

    result = query.first()
    if not result:
        raise exception.FailoverSegmentNotFound(id=segment_uuid)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def failover_segment_get_by_name(context, name):
    query = model_query(context, models.FailoverSegment).filter_by(name=name)

    result = query.first()
    if not result:
        raise exception.FailoverSegmentNotFoundByName(segment_name=name)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def failover_segment_create(context, values):

    segment = models.FailoverSegment()
    segment.update(values)
    try:
        segment.save(session=context.session)
    except db_exc.DBDuplicateEntry:
        raise exception.FailoverSegmentExists(name=segment.name)

    return _failover_segment_get_by_uuid(context, segment.uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def failover_segment_update(context, segment_uuid, values):
    segment = _failover_segment_get_by_uuid(context, segment_uuid)

    segment.update(values)
    try:
        segment.save(session=context.session)
    except db_exc.DBDuplicateEntry:
        raise exception.FailoverSegmentExists(name=values.get('name'))

    return _failover_segment_get_by_uuid(context, segment.uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def failover_segment_delete(context, segment_uuid):
    count = model_query(context, models.FailoverSegment
                        ).filter_by(uuid=segment_uuid
                                    ).soft_delete(synchronize_session=False)

    if count == 0:
        raise exception.FailoverSegmentNotFound(id=segment_uuid)

    model_query(context, models.Host).filter_by(
        failover_segment_id=segment_uuid).soft_delete(
        synchronize_session=False)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def is_failover_segment_under_recovery(context, failover_segment_id,
                                       filters=None):
    filters = filters or {}

    # get all hosts against the failover_segment
    inner_select = model_query(
        context, models.Host, (models.Host.uuid,)).filter(
            models.Host.failover_segment_id == failover_segment_id)

    # check if any host has notification status as new, running or error
    query = model_query(context, models.Notification,
                        (func.count(models.Notification.id),))
    if 'status' in filters:
        status = filters['status']
        if isinstance(status, (list, tuple, set, frozenset)):
            column_attr = getattr(models.Notification, 'status')
            query = query.filter(column_attr.in_(status))
        else:
            query = query.filter(models.Notification.status == status)

    query = query.filter(
        models.Notification.source_host_uuid.in_(inner_select.subquery()))

    return query.first()[0] > 0


# db apis for host


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def host_get_all_by_filters(
        context, filters=None, sort_keys=None,
        sort_dirs=None, limit=None, marker=None):

    # NOTE(Dinesh_Bhor): If the limit is 0 there is no point in even going
    # to the database since nothing is going to be returned anyway.
    if limit == 0:
        return []

    sort_keys, sort_dirs = _process_sort_params(sort_keys,
                                                sort_dirs)

    filters = filters or {}
    query = model_query(context,
                        models.Host).options(joinedload('failover_segment'))

    if 'failover_segment_id' in filters:
        query = query.filter(models.Host.failover_segment_id == filters[
            'failover_segment_id'])

    if 'type' in filters:
        query = query.filter(models.Host.type == filters['type'])

    if 'on_maintenance' in filters:
        query = query.filter(models.Host.on_maintenance == filters[
            'on_maintenance'])

    if 'reserved' in filters:
        query = query.filter(models.Host.reserved == filters['reserved'])

    marker_row = None
    if marker is not None:
        marker_row = model_query(context,
                                 models.Host
                                 ).filter_by(id=marker).first()
        if not marker_row:
            raise exception.MarkerNotFound(marker=marker)

    try:
        query = sqlalchemyutils.paginate_query(query, models.Host, limit,
                                               sort_keys,
                                               marker=marker_row,
                                               sort_dirs=sort_dirs)
    except db_exc.InvalidSortKey as e:
        raise exception.InvalidSortKey(e)

    return query.all()


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def host_get_by_uuid(context, host_uuid, segment_uuid=None):
    return _host_get_by_uuid(context, host_uuid, segment_uuid=segment_uuid)


def _host_get_by_uuid(context, host_uuid, segment_uuid=None):
    query = model_query(context, models.Host
                        ).filter_by(uuid=host_uuid
                                    ).options(joinedload('failover_segment'))
    if segment_uuid:
        query = query.filter_by(failover_segment_id=segment_uuid)

    result = query.first()

    if not result:
        if segment_uuid:
            raise exception.HostNotFoundUnderFailoverSegment(
                host_uuid=host_uuid, segment_uuid=segment_uuid)
        else:
            raise exception.HostNotFound(id=host_uuid)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def host_get_by_id(context, host_id):
    query = model_query(context, models.Host
                        ).filter_by(id=host_id
                                    ).options(joinedload('failover_segment'))

    result = query.first()
    if not result:
        raise exception.HostNotFound(id=host_id)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def host_get_by_name(context, name):
    query = model_query(context, models.Host
                        ).filter_by(name=name
                                    ).options(joinedload('failover_segment'))

    result = query.first()
    if not result:
        raise exception.HostNotFoundByName(host_name=name)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def host_create(context, values):
    host = models.Host()
    host.update(values)
    try:
        host.save(session=context.session)
    except db_exc.DBDuplicateEntry:
        raise exception.HostExists(name=host.name)

    return _host_get_by_uuid(context, host.uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def host_update(context, host_uuid, values):
    host = _host_get_by_uuid(context, host_uuid)

    host.update(values)
    try:
        host.save(session=context.session)
    except db_exc.DBDuplicateEntry:
        raise exception.HostExists(name=values.get('name'))

    return _host_get_by_uuid(context, host.uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def host_delete(context, host_uuid):

    count = model_query(context, models.Host
                        ).filter_by(uuid=host_uuid
                                    ).soft_delete(synchronize_session=False)

    if count == 0:
        raise exception.HostNotFound(id=host_uuid)


# db apis for notifications


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def notifications_get_all_by_filters(
        context, filters=None, sort_keys=None,
        sort_dirs=None, limit=None, marker=None):

    # NOTE(Dinesh_Bhor): If the limit is 0 there is no point in even going
    # to the database since nothing is going to be returned anyway.
    if limit == 0:
        return []

    sort_keys, sort_dirs = _process_sort_params(sort_keys,
                                                sort_dirs)

    filters = filters or {}
    query = model_query(context, models.Notification)

    if 'source_host_uuid' in filters:
        query = query.filter(models.Notification.source_host_uuid == filters[
            'source_host_uuid'])

    if 'type' in filters:
        query = query.filter(models.Notification.type == filters['type'])

    if 'status' in filters:
        status = filters['status']
        if isinstance(status, (list, tuple, set, frozenset)):
            column_attr = getattr(models.Notification, 'status')
            query = query.filter(column_attr.in_(status))
        else:
            query = query.filter(models.Notification.status == status)

    if 'generated-since' in filters:
        generated_since = timeutils.normalize_time(filters['generated-since'])
        query = query.filter(
            models.Notification.generated_time >= generated_since)

    marker_row = None
    if marker is not None:
        marker_row = model_query(context,
                                 models.Notification
                                 ).filter_by(id=marker).first()
        if not marker_row:
            raise exception.MarkerNotFound(marker=marker)

    try:
        query = sqlalchemyutils.paginate_query(query, models.Notification,
                                               limit,
                                               sort_keys,
                                               marker=marker_row,
                                               sort_dirs=sort_dirs)
    except db_exc.InvalidSortKey as err:
        raise exception.InvalidSortKey(err)

    return query.all()


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def notification_get_by_uuid(context, notification_uuid):
    return _notification_get_by_uuid(context, notification_uuid)


def _notification_get_by_uuid(context, notification_uuid):
    query = model_query(context, models.Notification
                        ).filter_by(notification_uuid=notification_uuid
                                    )

    result = query.first()
    if not result:
        raise exception.NotificationNotFound(id=notification_uuid)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.reader
def notification_get_by_id(context, notification_id):
    query = model_query(context, models.Notification
                        ).filter_by(id=notification_id
                                    )

    result = query.first()
    if not result:
        raise exception.NotificationNotFound(id=notification_id)

    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def notification_create(context, values):
    notification = models.Notification()
    notification.update(values)

    notification.save(session=context.session)

    return _notification_get_by_uuid(context, notification.notification_uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def notification_update(context, notification_uuid, values):
    notification = _notification_get_by_uuid(context, notification_uuid)

    notification.update(values)

    notification.save(session=context.session)

    return _notification_get_by_uuid(context, notification.notification_uuid)


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def notification_delete(context, notification_uuid):

    count = model_query(context, models.Notification
                        ).filter_by(notification_uuid=notification_uuid
                                    ).soft_delete(synchronize_session=False)

    if count == 0:
        raise exception.NotificationNotFound(id=notification_uuid)


class DeleteFromSelect(sa_sql.expression.UpdateBase):
    def __init__(self, table, select, column):
        self.table = table
        self.select = select
        self.column = column


# NOTE(pooja_jadhav): MySQL doesn't yet support subquery with
# 'LIMIT & IN/ALL/ANY/SOME' We need work around this with nesting select.
@compiles(DeleteFromSelect)
def visit_delete_from_select(element, compiler, **kw):
    return "DELETE FROM %s WHERE %s in (SELECT T1.%s FROM (%s) as T1)" % (
        compiler.process(element.table, asfrom=True),
        compiler.process(element.column),
        element.column.name,
        compiler.process(element.select))


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@main_context_manager.writer
def purge_deleted_rows(context, age_in_days, max_rows):
    """Purges soft deleted rows

    Deleted rows get purged from hosts and segment tables based on
    deleted_at column. As notifications table doesn't delete any of
    the notification records so rows get purged from notifications
    based on last updated_at and status column.
    """
    engine = get_engine()
    conn = engine.connect()
    metadata = MetaData()
    metadata.reflect(engine)
    deleted_age = timeutils.utcnow() - datetime.timedelta(days=age_in_days)
    total_rows_purged = 0
    for table in reversed(metadata.sorted_tables):
        if 'deleted' not in table.columns.keys():
            continue
        LOG.info('Purging deleted rows older than %(age_in_days)d day(s) '
                 'from table %(tbl)s',
            {'age_in_days': age_in_days, 'tbl': table})
        column = table.c.id
        updated_at_column = table.c.updated_at
        deleted_at_column = table.c.deleted_at

        if table.name == 'notifications':
            status_column = table.c.status
            query_delete = sql.select([column]).where(
                and_(updated_at_column < deleted_age, or_(
                    status_column == 'finished', status_column == 'failed',
                    status_column == 'ignored'))).order_by(status_column)
        else:
            query_delete = sql.select(
                [column], deleted_at_column < deleted_age).order_by(
                deleted_at_column)

        if max_rows > 0:
            query_delete = query_delete.limit(max_rows - total_rows_purged)

        delete_statement = DeleteFromSelect(table, query_delete, column)

        result = conn.execute(delete_statement)

        rows = result.rowcount
        LOG.info('Deleted %(rows)d row(s) from table %(tbl)s',
                 {'rows': rows, 'tbl': table})

        total_rows_purged += rows
        if max_rows > 0 and total_rows_purged == max_rows:
            break

    LOG.info('Total deleted rows are %(rows)d', {'rows': total_rows_purged})



# Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_log import log as logging
from urllib import parse as urlparse
import webob

import masakari.conf
from masakari import exception
from masakari.i18n import _
from masakari import utils


CONF = masakari.conf.CONF

LOG = logging.getLogger(__name__)


def remove_trailing_version_from_href(href):
    """Removes the api version from the href.

    Given: 'http://www.masakari.com/ha/v1.1'
    Returns: 'http://www.masakari.com/ha'

    Given: 'http://www.masakari.com/v1.1'
    Returns: 'http://www.masakari.com'

    """
    parsed_url = urlparse.urlsplit(href)
    url_parts = parsed_url.path.rsplit('/', 1)

    # NOTE: this should match vX.X or vX
    expression = re.compile(r'^v([0-9]+|[0-9]+\.[0-9]+)(/.*|$)')
    if not expression.match(url_parts.pop()):
        LOG.debug('href %s does not contain version', href)
        raise ValueError(_('href %s does not contain version') % href)

    new_path = url_join(*url_parts)
    parsed_url = list(parsed_url)
    parsed_url[2] = new_path
    return urlparse.urlunsplit(parsed_url)


def url_join(*parts):
    """Convenience method for joining parts of a URL

    Any leading and trailing '/' characters are removed, and the parts joined
    together with '/' as a separator. If last element of 'parts' is an empty
    string, the returned URL will have a trailing slash.
    """
    parts = parts or [""]
    clean_parts = [part.strip("/") for part in parts if part]
    if not parts[-1]:
        # Empty last element should add a trailing slash
        clean_parts.append("")
    return "/".join(clean_parts)


class ViewBuilder(object):
    """Model API responses as dictionaries."""

    def _get_project_id(self, request):
        """Get project id from request url if present or empty string
        otherwise
        """
        project_id = request.environ["masakari.context"].project_id
        if project_id in request.url:
            return project_id
        return ''

    def _get_links(self, request, identifier, collection_name):
        return [
            {
                "rel": "self",
                "href": self._get_href_link(request, identifier,
                                            collection_name),
            },
            {
                "rel": "bookmark",
                "href": self._get_bookmark_link(request,
                                                identifier,
                                                collection_name),
            }
        ]

    def _get_next_link(self, request, identifier, collection_name):
        """Return href string with proper limit and marker params."""
        params = request.params.copy()
        params["marker"] = identifier
        prefix = self._update_masakari_link_prefix(request.application_url)
        url = url_join(prefix,
                       self._get_project_id(request),
                       collection_name)
        return "%s?%s" % (url, urlparse.urlencode(params))

    def _get_href_link(self, request, identifier, collection_name):
        """Return an href string pointing to this object."""
        prefix = self._update_masakari_link_prefix(request.application_url)
        return url_join(prefix,
                        self._get_project_id(request),
                        collection_name,
                        str(identifier))

    def _get_bookmark_link(self, request, identifier, collection_name):
        """Create a URL that refers to a specific resource."""
        base_url = remove_trailing_version_from_href(request.application_url)
        base_url = self._update_masakari_link_prefix(base_url)
        return url_join(base_url,
                        self._get_project_id(request),
                        collection_name,
                        str(identifier))

    def _get_collection_links(self,
                              request,
                              items,
                              collection_name,
                              id_key="uuid"):
        """Retrieve 'next' link, if applicable. This is included if:
        1) 'limit' param is specified and equals the number of items.
        2) 'limit' param is specified but it exceeds CONF.osapi_max_limit,
        in this case the number of items is CONF.osapi_max_limit.
        3) 'limit' param is NOT specified but the number of items is
        CONF.osapi_max_limit.
        """
        links = []
        max_items = min(
            int(request.params.get("limit", CONF.osapi_max_limit)),
            CONF.osapi_max_limit)
        if max_items and max_items == len(items):
            last_item = items[-1]
            if id_key in last_item:
                last_item_id = last_item[id_key]
            elif 'id' in last_item:
                last_item_id = last_item["id"]
            else:
                last_item_id = last_item["flavorid"]
            links.append({
                "rel": "next",
                "href": self._get_next_link(request,
                                            last_item_id,
                                            collection_name),
            })
        return links

    def _update_link_prefix(self, orig_url, prefix):
        if not prefix:
            return orig_url
        url_parts = list(urlparse.urlsplit(orig_url))
        prefix_parts = list(urlparse.urlsplit(prefix))
        url_parts[0:2] = prefix_parts[0:2]
        url_parts[2] = prefix_parts[2] + url_parts[2]
        return urlparse.urlunsplit(url_parts).rstrip('/')

    def _update_masakari_link_prefix(self, orig_url):
        return self._update_link_prefix(orig_url,
                                        CONF.osapi_masakari_link_prefix)


def _get_int_param(request, param):
    """Extract integer param from request or fail."""
    try:
        int_param = utils.validate_integer(request.GET[param], param,
                                           min_value=0)
    except exception.InvalidInput as e:
        raise webob.exc.HTTPBadRequest(explanation=e.format_message())
    return int_param


def _get_marker_param(request):
    """Extract marker id from request or fail."""
    return request.GET['marker']


def get_pagination_params(request):
    """Return marker, limit tuple from request.

    :param request: `wsgi.Request` possibly containing 'marker' and 'limit'
                    GET variables. 'marker' is the id of the last element
                    the client has seen, and 'limit' is the maximum number
                    of items to return. If 'limit' is not specified, 0, or
                    > max_limit, we default to max_limit. Negative values
                    for either marker or limit will cause
                    exc.HTTPBadRequest() exceptions to be raised.

    """
    params = {}
    if 'limit' in request.GET:
        params['limit'] = _get_int_param(request, 'limit')
    if 'page_size' in request.GET:
        params['page_size'] = _get_int_param(request, 'page_size')
    if 'marker' in request.GET:
        params['marker'] = _get_marker_param(request)
    return params


def get_limit_and_marker(request, max_limit=CONF.osapi_max_limit):
    """get limited parameter from request."""
    params = get_pagination_params(request)
    limit = params.get('limit', max_limit)
    limit = min(max_limit, limit)
    marker = params.get('marker')

    return limit, marker


def get_sort_params(input_params, default_key='created_at',
                    default_dir='desc'):
    """Retrieves sort keys/directions parameters.

    Processes the parameters to create a list of sort keys and sort directions
    that correspond to the 'sort_key' and 'sort_dir' parameter values. These
    sorting parameters can be specified multiple times in order to generate
    the list of sort keys and directions.

    The input parameters are not modified.

    :param input_params: webob.multidict of request parameters (from
                         masakari.wsgi.Request.params)
    :param default_key: default sort key value, added to the list if no
                        'sort_key' parameters are supplied
    :param default_dir: default sort dir value, added to the list if no
                        'sort_dir' parameters are supplied
    :returns: list of sort keys, list of sort dirs
    """
    params = input_params.copy()
    sort_keys = []
    sort_dirs = []

    while 'sort_key' in params:
        sort_keys.append(params.pop('sort_key').strip())
    while 'sort_dir' in params:
        sort_dirs.append(params.pop('sort_dir').strip())
    if len(sort_keys) == 0 and default_key:
        sort_keys.append(default_key)
    if len(sort_dirs) == 0 and default_dir:
        sort_dirs.append(default_dir)
    return sort_keys, sort_dirs



# Copyright (c) 2018 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from masakari.notifications.objects import base
from masakari.objects import base as masakari_base
from masakari.objects import fields


@masakari_base.MasakariObjectRegistry.register_notification
class SegmentApiPayloadBase(base.NotificationPayloadBase):
    SCHEMA = {
        'id': ('segment', 'id'),
        'uuid': ('segment', 'uuid'),
        'name': ('segment', 'name'),
        'service_type': ('segment', 'service_type'),
        'description': ('segment', 'description'),
        'recovery_method': ('segment', 'recovery_method'),
        'enabled': ('segment', 'enabled'),
    }
    # Version 1.0: Initial version
    # Version 1.1: Add 'enabled' field
    VERSION = '1.1'
    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(),
        'name': fields.StringField(),
        'service_type': fields.StringField(),
        'description': fields.StringField(nullable=True),
        'recovery_method': fields.FailoverSegmentRecoveryMethodField(),
        'enabled': fields.BooleanField(),
        }

    def __init__(self, segment, **kwargs):
        super(SegmentApiPayloadBase, self).__init__(**kwargs)
        self.populate_schema(segment=segment)


@masakari_base.MasakariObjectRegistry.register_notification
class SegmentApiPayload(SegmentApiPayloadBase):
    # No SCHEMA as all the additional fields are calculated

    VERSION = '1.1'
    fields = {
        'fault': fields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, segment, fault, **kwargs):
        super(SegmentApiPayload, self).__init__(
            segment=segment,
            fault=fault,
            **kwargs)


@masakari_base.MasakariObjectRegistry.register_notification
class HostApiPayloadBase(base.NotificationPayloadBase):
    SCHEMA = {
        'id': ('host', 'id'),
        'uuid': ('host', 'uuid'),
        'name': ('host', 'name'),
        'failover_segment': ('host', 'failover_segment'),
        'type': ('host', 'type'),
        'reserved': ('host', 'reserved'),
        'control_attributes': ('host', 'control_attributes'),
        'on_maintenance': ('host', 'on_maintenance'),
    }
    # Version 1.0: Initial version
    # Version 1.1: Removed 'failover_segment_id' parameter
    VERSION = '1.1'
    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(),
        'name': fields.StringField(),
        'failover_segment': fields.ObjectField('FailoverSegment'),
        'type': fields.StringField(),
        'reserved': fields.BooleanField(),
        'control_attributes': fields.StringField(),
        'on_maintenance': fields.BooleanField(),
        }

    def __init__(self, host, **kwargs):
        super(HostApiPayloadBase, self).__init__(**kwargs)
        self.populate_schema(host=host)


@masakari_base.MasakariObjectRegistry.register_notification
class HostApiPayload(HostApiPayloadBase):
    # No SCHEMA as all the additional fields are calculated

    VERSION = '1.0'
    fields = {
        'fault': fields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, host, fault, **kwargs):
        super(HostApiPayload, self).__init__(
            host=host,
            fault=fault,
            **kwargs)


@masakari_base.MasakariObjectRegistry.register_notification
class NotificationApiPayloadBase(base.NotificationPayloadBase):
    SCHEMA = {
        'id': ('notification', 'id'),
        'notification_uuid': ('notification', 'notification_uuid'),
        'generated_time': ('notification', 'generated_time'),
        'source_host_uuid': ('notification', 'source_host_uuid'),
        'type': ('notification', 'type'),
        'payload': ('notification', 'payload'),
        'status': ('notification', 'status'),
    }
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'id': fields.IntegerField(),
        'notification_uuid': fields.UUIDField(),
        'generated_time': fields.DateTimeField(),
        'source_host_uuid': fields.UUIDField(),
        'type': fields.NotificationTypeField(),
        'payload': fields.DictOfStringsField(),
        'status': fields.NotificationStatusField(),
        }

    def __init__(self, notification, **kwargs):
        super(NotificationApiPayloadBase, self).__init__(**kwargs)
        self.populate_schema(notification=notification)


@masakari_base.MasakariObjectRegistry.register_notification
class NotificationApiPayload(NotificationApiPayloadBase):
    # No SCHEMA as all the additional fields are calculated

    VERSION = '1.0'
    fields = {
        'fault': fields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, notification, fault, **kwargs):
        super(NotificationApiPayload, self).__init__(
            notification=notification,
            fault=fault,
            **kwargs)


@base.notification_sample('create-segment-start.json')
@base.notification_sample('create-segment-end.json')
@base.notification_sample('update-segment-start.json')
@base.notification_sample('update-segment-end.json')
@base.notification_sample('delete-segment-start.json')
@base.notification_sample('delete-segment-end.json')
@masakari_base.MasakariObjectRegistry.register_notification
class SegmentApiNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('SegmentApiPayload')
    }


@base.notification_sample('create-host-start.json')
@base.notification_sample('create-host-end.json')
@base.notification_sample('update-host-start.json')
@base.notification_sample('update-host-end.json')
@base.notification_sample('delete-host-start.json')
@base.notification_sample('delete-host-end.json')
@masakari_base.MasakariObjectRegistry.register_notification
class HostApiNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('HostApiPayload')
    }


@base.notification_sample('create-notification-start.json')
@base.notification_sample('create-notification-end.json')
@base.notification_sample('process-notification-start.json')
@base.notification_sample('process-notification-end.json')
@base.notification_sample('process-notification-error.json')
@masakari_base.MasakariObjectRegistry.register_notification
class NotificationApiNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('NotificationApiPayload')
    }



# Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
WSGI middleware for OpenStack API controllers.
"""

from oslo_log import log as logging
import routes
import stevedore
import webob.dec
import webob.exc

from masakari.api.openstack import wsgi
import masakari.conf
from masakari.i18n import translate
from masakari import utils
from masakari import wsgi as base_wsgi


LOG = logging.getLogger(__name__)
CONF = masakari.conf.CONF


class FaultWrapper(base_wsgi.Middleware):
    """Calls down the middleware stack, making exceptions into faults."""

    _status_to_type = {}

    @staticmethod
    def status_to_type(status):
        if not FaultWrapper._status_to_type:
            for clazz in utils.walk_class_hierarchy(webob.exc.HTTPError):
                FaultWrapper._status_to_type[clazz.code] = clazz
        return FaultWrapper._status_to_type.get(
            status, webob.exc.HTTPInternalServerError)()

    def _error(self, inner, req):
        LOG.exception("Caught error: %s", str(inner))

        safe = getattr(inner, 'safe', False)
        headers = getattr(inner, 'headers', None)
        status = getattr(inner, 'code', 500)
        if status is None:
            status = 500

        msg_dict = dict(url=req.url, status=status)
        LOG.info("%(url)s returned with HTTP %(status)d", msg_dict)
        outer = self.status_to_type(status)
        if headers:
            outer.headers = headers

        if safe:
            user_locale = req.best_match_language()
            inner_msg = translate(inner.message, user_locale)
            outer.explanation = '%s: %s' % (inner.__class__.__name__,
                                            inner_msg)

        return wsgi.Fault(outer)

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        try:
            return req.get_response(self.application)
        except Exception as ex:
            return self._error(ex, req)


class APIMapper(routes.Mapper):
    def routematch(self, url=None, environ=None):
        if url == "":
            result = self._match("", environ)
            return result[0], result[1]
        return routes.Mapper.routematch(self, url, environ)

    def connect(self, *args, **kargs):
        kargs.setdefault('requirements', {})
        if not kargs['requirements'].get('format'):
            kargs['requirements']['format'] = 'json|xml'
        return routes.Mapper.connect(self, *args, **kargs)


class ProjectMapper(APIMapper):
    def resource(self, member_name, collection_name, **kwargs):
        # NOTE(abhishekk): project_id parameter is only valid if its hex
        # or hex + dashes (note, integers are a subset of this). This
        # is required to hand our overlaping routes issues.
        project_id_regex = r'[0-9a-f\-]+'
        if CONF.osapi_v1.project_id_regex:
            project_id_regex = CONF.osapi_v1.project_id_regex

        project_id_token = '{project_id:%s}' % project_id_regex
        if 'parent_resource' not in kwargs:
            kwargs['path_prefix'] = '%s/' % project_id_token
        else:
            parent_resource = kwargs['parent_resource']
            p_collection = parent_resource['collection_name']
            p_member = parent_resource['member_name']
            kwargs['path_prefix'] = '%s/%s/:%s_id' % (
                project_id_token,
                p_collection,
                p_member)
        routes.Mapper.resource(
            self,
            member_name,
            collection_name,
            **kwargs)

        # while we are in transition mode, create additional routes
        # for the resource that do not include project_id.
        if 'parent_resource' not in kwargs:
            del kwargs['path_prefix']
        else:
            parent_resource = kwargs['parent_resource']
            p_collection = parent_resource['collection_name']
            p_member = parent_resource['member_name']
            kwargs['path_prefix'] = '%s/:%s_id' % (p_collection,
                                                   p_member)
        routes.Mapper.resource(self, member_name,
                               collection_name,
                               **kwargs)


class PlainMapper(APIMapper):
    def resource(self, member_name, collection_name, **kwargs):
        if 'parent_resource' in kwargs:
            parent_resource = kwargs['parent_resource']
            p_collection = parent_resource['collection_name']
            p_member = parent_resource['member_name']
            kwargs['path_prefix'] = '%s/:%s_id' % (p_collection, p_member)
        routes.Mapper.resource(self, member_name,
                               collection_name,
                               **kwargs)


class APIRouterV1(base_wsgi.Router):
    """Routes requests on the OpenStack v1 API to the appropriate controller
    and method.
    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Simple paste factory

        :class:`masakari.wsgi.Router` doesn't have one.
        """
        return cls()

    @staticmethod
    def api_extension_namespace():
        return 'masakari.api.v1.extensions'

    def __init__(self, init_only=None):
        def _check_load_extension(ext):
            return self._register_extension(ext)

        self.api_extension_manager = stevedore.enabled.EnabledExtensionManager(
            namespace=self.api_extension_namespace(),
            check_func=_check_load_extension,
            invoke_on_load=True,
            invoke_kwds={"extension_info": self.loaded_extension_info})

        mapper = ProjectMapper()

        self.resources = {}

        if list(self.api_extension_manager):
            self._register_resources_check_inherits(mapper)
            self.api_extension_manager.map(self._register_controllers)

        LOG.info("Loaded extensions: %s",
                 sorted(self.loaded_extension_info.get_extensions().keys()))
        super(APIRouterV1, self).__init__(mapper)

    def _register_resources_list(self, ext_list, mapper):
        for ext in ext_list:
            self._register_resources(ext, mapper)

    def _register_resources_check_inherits(self, mapper):
        ext_has_inherits = []
        ext_no_inherits = []

        for ext in self.api_extension_manager:
            for resource in ext.obj.get_resources():
                if resource.inherits:
                    ext_has_inherits.append(ext)
                    break
            else:
                ext_no_inherits.append(ext)

        self._register_resources_list(ext_no_inherits, mapper)
        self._register_resources_list(ext_has_inherits, mapper)

    @property
    def loaded_extension_info(self):
        raise NotImplementedError()

    def _register_extension(self, ext):
        raise NotImplementedError()

    def _register_resources(self, ext, mapper):
        """Register resources defined by the extensions

        Extensions define what resources they want to add through a
        get_resources function
        """

        handler = ext.obj
        LOG.debug("Running _register_resources on %s", ext.obj)

        for resource in handler.get_resources():
            LOG.debug('Extended resource: %s', resource.collection)

            inherits = None
            if resource.inherits:
                inherits = self.resources.get(resource.inherits)
                if not resource.controller:
                    resource.controller = inherits.controller
            wsgi_resource = wsgi.ResourceV1(resource.controller,
                                            inherits=inherits)
            self.resources[resource.collection] = wsgi_resource
            kargs = dict(
                controller=wsgi_resource,
                collection=resource.collection_actions,
                member=resource.member_actions)

            if resource.parent:
                kargs['parent_resource'] = resource.parent

            # non core-API plugins use the collection name as the
            # member name, but the core-API plugins use the
            # singular/plural convention for member/collection names
            if resource.member_name:
                member_name = resource.member_name
            else:
                member_name = resource.collection
            mapper.resource(member_name, resource.collection,
                            **kargs)

            if resource.custom_routes_fn:
                resource.custom_routes_fn(mapper, wsgi_resource)

    def _register_controllers(self, ext):
        """Register controllers defined by the extensions

        Extensions define what resources they want to add through
        a get_controller_extensions function
        """

        handler = ext.obj
        LOG.debug("Running _register_controllers on %s", ext.obj)

        for extension in handler.get_controller_extensions():
            ext_name = extension.extension.name
            collection = extension.collection
            controller = extension.controller

            if collection not in self.resources:
                LOG.warning('Extension %(ext_name)s: Cannot extend '
                            'resource %(collection)s: No such resource',
                            {'ext_name': ext_name, 'collection': collection})
                continue

            LOG.debug('Extension %(ext_name)s extending resource: '
                      '%(collection)s',
                      {'ext_name': ext_name, 'collection': collection})

            resource = self.resources[collection]
            resource.register_actions(controller)
            resource.register_extensions(controller)



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the notifications api."""

import copy
from http import client as http
from unittest import mock

import ddt
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from webob import exc

from masakari.api.openstack.ha import notifications
from masakari.engine import rpcapi as engine_rpcapi
from masakari import exception
from masakari.ha import api as ha_api
from masakari.objects import base as obj_base
from masakari.objects import fields
from masakari.objects import notification as notification_obj
from masakari import test
from masakari.tests.unit.api.openstack import fakes
from masakari.tests.unit.objects import test_objects
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)
OPTIONAL = ['recovery_workflow_details']


def _make_notification_obj(notification_dict):
    return notification_obj.Notification(**notification_dict)


def _make_notification_progress_details_obj(progress_details):
    return notification_obj.NotificationProgressDetails(**progress_details)


def _make_notifications_list(notifications_list):
    return notification_obj.Notification(objects=[
        _make_notification_obj(a) for a in notifications_list])


NOTIFICATION_DATA = {"type": "VM", "id": 1,
                     "payload":
                         {'event': 'STOPPED', 'host_status': 'NORMAL',
                          'cluster_status': 'ONLINE'},
                     "source_host_uuid": uuidsentinel.fake_host,
                     "generated_time": NOW,
                     "status": "running",
                     "notification_uuid": uuidsentinel.fake_notification,
                     "created_at": NOW,
                     "updated_at": None,
                     "deleted_at": None,
                     "deleted": 0
                     }

NOTIFICATION = _make_notification_obj(NOTIFICATION_DATA)

RECOVERY_DETAILS = {"progress": 1.0,
                    "state": "SUCCESS",
                    "name": "StopInstanceTask",
                    "progress_details": [
                        {"timestamp": "2019-03-07 13:54:28",
                         "message": "Stopping instance",
                         "progress": "0.0"},
                    ]}

NOTI_DATA_WITH_DETAILS = copy.deepcopy(NOTIFICATION_DATA)

NOTIFICATION_WITH_PROGRESS_DETAILS = _make_notification_obj(
    NOTI_DATA_WITH_DETAILS)
RECOVERY_OBJ = _make_notification_progress_details_obj(RECOVERY_DETAILS)
NOTIFICATION_WITH_PROGRESS_DETAILS.recovery_workflow_details = [RECOVERY_OBJ]

NOTIFICATION_LIST = [
    {"type": "VM", "id": 1, "payload": {'event': 'STOPPED',
                                        'host_status': 'NORMAL',
                                        'cluster_status': 'ONLINE'},
     "source_host_uuid": uuidsentinel.fake_host, "generated_time": NOW,
     "status": "running", "notification_uuid": uuidsentinel.fake_notification,
     "created_at": NOW, "updated_at": None, "deleted_at": None, "deleted": 0},

    {"type": "PROCESS", "id": 2, "payload": {'event': 'STOPPED',
                                             'process_name': 'fake_process'},
     "source_host_uuid": uuidsentinel.fake_host1, "generated_time": NOW,
     "status": "running", "notification_uuid": uuidsentinel.fake_notification1,
     "created_at": NOW, "updated_at": None, "deleted_at": None, "deleted": 0},
]

NOTIFICATION_LIST = _make_notifications_list(NOTIFICATION_LIST)


@ddt.ddt
class NotificationTestCase(test.TestCase):
    """Test Case for notifications api."""

    bad_request = exception.ValidationError

    @mock.patch.object(engine_rpcapi, 'EngineAPI')
    def setUp(self, mock_rpc):
        super(NotificationTestCase, self).setUp()
        self.controller = notifications.NotificationsController()
        self.req = fakes.HTTPRequest.blank('/v1/notifications',
                                           use_admin_context=True)
        self.context = self.req.environ['masakari.context']

    @property
    def app(self):
        return fakes.wsgi_app_v1(init_only='os-hosts')

    def _assert_notification_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The notifications objects were not equal")

    @mock.patch.object(ha_api.NotificationAPI, 'get_all')
    def test_index(self, mock_get_all):

        mock_get_all.return_value = NOTIFICATION_LIST

        result = self.controller.index(self.req)
        result = result['notifications']
        self._assert_notification_data(NOTIFICATION_LIST,
                                       _make_notifications_list(result))

    @ddt.data(
        # limit negative
        "limit=-1",

        # invalid sort key
        "sort_key=abcd",

        # invalid sort dir
        "sort_dir=abcd")
    def test_index_invalid(self, param):
        req = fakes.HTTPRequest.blank("/v1/notifications?%s" % param,
                                      use_admin_context=True)

        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req)

    @mock.patch.object(ha_api.NotificationAPI, 'get_all')
    def test_index_marker_not_found(self, mock_get_all):
        fake_request = fakes.HTTPRequest.blank('/v1/notifications?marker=1234',
                                               use_admin_context=True)
        mock_get_all.side_effect = exception.MarkerNotFound(marker="1234")
        self.assertRaises(exc.HTTPBadRequest, self.controller.index,
                          fake_request)

    def test_index_invalid_generated_since(self):

        req = fakes.HTTPRequest.blank('/v1/notifications?generated-since=abcd',
                                      use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req)

    @mock.patch.object(ha_api.NotificationAPI, 'get_all')
    def test_index_valid_generated_since(self, mock_get_all):
        url = '/v1/notifications?generated-since=%s' % str(NOW)
        req = fakes.HTTPRequest.blank(url, use_admin_context=True)
        mock_get_all.return_value = NOTIFICATION_LIST
        result = self.controller.index(req)
        result = result['notifications']
        self._assert_notification_data(NOTIFICATION_LIST,
                                       _make_notifications_list(result))

    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create(self, mock_create):

        mock_create.return_value = NOTIFICATION
        result = self.controller.create(self.req, body={
            "notification": {
                "hostname": "fake_host",
                "payload": {
                    "instance_uuid": uuidsentinel.instance_uuid,
                    "vir_domain_event": "STOPPED_FAILED",
                    "event": "LIFECYCLE"
                },
                "type": "VM",
                "generated_time": "2016-09-13T09:11:21.656788"}})
        result = result['notification']
        test_objects.compare_obj(self, result, NOTIFICATION_DATA,
                                 allow_missing=OPTIONAL)

    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create_process_notification(self, mock_create):
        mock_create.return_value = NOTIFICATION
        result = self.controller.create(self.req, body={
            "notification": {
                "hostname": "fake_host",
                "payload": {
                    "process_name": "nova-compute",
                    "event": "STOPPED"
                },
                "type": "PROCESS",
                "generated_time": "2016-09-13T09:11:21.656788"}})
        result = result['notification']
        test_objects.compare_obj(self, result, NOTIFICATION_DATA,
                                 allow_missing=OPTIONAL)

    @mock.patch('masakari.rpc.get_client')
    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create_success_with_201_response_code(
        self, mock_client, mock_create):
        body = {
            "notification": {
                "hostname": "fake_host",
                "payload": {
                    "instance_uuid": uuidsentinel.instance_uuid,
                    "vir_domain_event": "STOPPED_FAILED",
                    "event": "LIFECYCLE"
                },
                "type": "VM",
                "generated_time": NOW
            }
        }
        fake_req = self.req
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = 'POST'
        fake_req.body = jsonutils.dump_as_bytes(body)
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.ACCEPTED, resp.status_code)

    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create_host_not_found(self, mock_create):
        body = {
            "notification": {
                "hostname": "fake_host",
                "payload": {
                    "instance_uuid": uuidsentinel.instance_uuid,
                    "vir_domain_event": "STOPPED_FAILED",
                    "event": "LIFECYCLE"
                },
                "type": "VM",
                "generated_time": "2016-09-13T09:11:21.656788"
            }
        }
        mock_create.side_effect = exception.HostNotFoundByName(
            host_name="fake_host")
        self.assertRaises(exc.HTTPBadRequest, self.controller.create,
                          self.req, body=body)

    @ddt.data(
        # invalid type
        {"body": {
            "notification": {"hostname": "fake_host",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "Fake",
                             "generated_time": "2016-09-13T09:11:21.656788"}}},

        # no notification in body
        {"body": {"hostname": "fake_host",
                  "payload": {"event": "STOPPED",
                              "host_status": "NORMAL",
                              "cluster_status": "ONLINE"},
                  "type": "VM",
                  "generated_time": "2016-09-13T09:11:21.656788"}},

        # no payload
        {"body": {"notification": {"hostname": "fake_host",
                                   "type": "VM",
                                   "generated_time":
                                       "2016-09-13T09:11:21.656788"}}},

        # no hostname
        {"body": {"notification": {"payload": {"event": "STOPPED",
                                               "host_status": "NORMAL",
                                               "cluster_status": "ONLINE"},
                                   "type": "VM",
                                   "generated_time":
                                       "2016-09-13T09:11:21.656788"}}},

        # no type
        {"body": {"notification": {"hostname": "fake_host",
                                   "payload": {"event": "STOPPED",
                                               "host_status": "NORMAL",
                                               "cluster_status": "ONLINE"},
                                   "generated_time":
                                       "2016-09-13T09:11:21.656788"}}},

        # no generated time
        {"body": {"notification": {"hostname": "fake_host",
                                   "payload": {"event": "STOPPED",
                                               "host_status": "NORMAL",
                                               "cluster_status": "ONLINE"},
                                   "type": "VM",
                                   }}},

        # hostname too long
        {"body": {
            "notification": {"hostname": "fake_host" * 255,
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "VM",
                             "generated_time": "2016-09-13T09:11:21.656788"}}},

        # extra invalid args
        {"body": {
            "notification": {"hostname": "fake_host",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "VM",
                             "generated_time": "2016-09-13T09:11:21.656788",
                             "invalid_extra": "non_expected_parameter"}}}
    )
    @ddt.unpack
    def test_create_failure(self, body):
        self.assertRaises(self.bad_request, self.controller.create,
                          self.req, body=body)

    @ddt.data(
        # invalid event for PROCESS type
        {"params": {"payload": {"event": "invalid",
                                "process_name": "nova-compute"},
                    "type": fields.NotificationType.PROCESS}},

        # invalid event for VM type
        {"params": {"payload": {"event": "invalid",
                                "host_status": fields.HostStatusType.NORMAL,
                          "cluster_status": fields.ClusterStatusType.ONLINE},
                    "type": fields.NotificationType.VM}},

        # invalid event for HOST_COMPUTE type
        {"params": {"payload": {"event": "invalid"},
                    "type": fields.NotificationType.COMPUTE_HOST}},

        # empty payload
        {"params": {"payload": {},
                    "type": fields.NotificationType.COMPUTE_HOST}},

        # empty process_name
        {"params": {"payload": {"event": fields.EventType.STOPPED,
                                "process_name": ""},
                    "type": fields.NotificationType.PROCESS}},

        # process_name too long value
        {"params": {"payload": {"event": fields.EventType.STOPPED,
                              "process_name": "a" * 4097},
                  "type": fields.NotificationType.PROCESS}},

        # process_name invalid data_type
        {"params": {"payload": {"event": fields.EventType.STOPPED,
                              "process_name": 123},
                  "type": fields.NotificationType.PROCESS}}
    )
    @ddt.unpack
    def test_create_with_invalid_payload(self, params):
        body = {
            "notification": {"hostname": "fake_host",
                             "generated_time": "2016-09-13T09:11:21.656788"
                             }
        }

        body['notification']['payload'] = params['payload']
        body['notification']['type'] = params['type']
        self.assertRaises(self.bad_request, self.controller.create,
                          self.req, body=body)

    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create_duplicate_notification(self, mock_create_notification):
        mock_create_notification.side_effect = exception.DuplicateNotification(
            type="COMPUTE_HOST")
        body = {
            "notification": {"hostname": "fake_host",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "COMPUTE_HOST",
                             "generated_time": str(NOW)}}
        self.assertRaises(exc.HTTPConflict, self.controller.create,
                          self.req, body=body)

    @mock.patch.object(ha_api.NotificationAPI, 'create_notification')
    def test_create_host_on_maintenance(self, mock_create_notification):
        mock_create_notification.side_effect = (
            exception.HostOnMaintenanceError(host_name="fake_host"))
        body = {
            "notification": {"hostname": "fake_host",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "COMPUTE_HOST",
                             "generated_time": str(NOW)}}
        self.assertRaises(exc.HTTPConflict, self.controller.create,
                          self.req, body=body)

    @mock.patch.object(ha_api.NotificationAPI, 'get_notification')
    def test_show(self, mock_get_notification):

        mock_get_notification.return_value = NOTIFICATION

        result = self.controller.show(self.req, uuidsentinel.fake_notification)
        result = result['notification']
        self._assert_notification_data(NOTIFICATION,
                                       _make_notification_obj(result))

    @mock.patch.object(ha_api.NotificationAPI, 'get_notification')
    def test_show_with_non_existing_uuid(self, mock_get_notification):

        mock_get_notification.side_effect = exception.NotificationNotFound(
            id="2")
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show, self.req, "2")

    @ddt.data('DELETE', 'PUT')
    @mock.patch('masakari.rpc.get_client')
    def test_delete_and_update_notification(self, method, mock_client):
        url = '/v1/notifications/%s' % uuidsentinel.fake_notification
        fake_req = fakes.HTTPRequest.blank(url, use_admin_context=True)
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = method
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.METHOD_NOT_ALLOWED, resp.status_code)


class NotificationCasePolicyNotAuthorized(test.NoDBTestCase):
    """Test Case for notifications non admin."""

    @mock.patch.object(engine_rpcapi, 'EngineAPI')
    def setUp(self, mock_rpc):
        super(NotificationCasePolicyNotAuthorized, self).setUp()
        self.controller = notifications.NotificationsController()
        self.req = fakes.HTTPRequest.blank('/v1/notifications')
        self.context = self.req.environ['masakari.context']

    def _check_rule(self, exc, rule_name):
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())

    def test_create_no_admin(self):
        rule_name = "os_masakari_api:notifications:create"
        self.policy.set_rules({rule_name: "project:non_fake"})
        body = {
            "notification": {"hostname": "fake_host",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "VM",
                             "generated_time": "2016-09-13T09:11:21.656788"}}
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.create,
                                self.req, body=body)
        self._check_rule(exc, rule_name)

    def test_show_no_admin(self):
        rule_name = "os_masakari_api:notifications:detail"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.show,
                                self.req, uuidsentinel.fake_notification)
        self._check_rule(exc, rule_name)

    def test_index_no_admin(self):
        rule_name = "os_masakari_api:notifications:index"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.index,
                                self.req)
        self._check_rule(exc, rule_name)


class NotificationV1_1_TestCase(NotificationTestCase):
    """Test Case for notifications api for 1.1 API"""
    api_version = '1.1'

    @mock.patch.object(engine_rpcapi, 'EngineAPI')
    def setUp(self, mock_rpc):
        super(NotificationV1_1_TestCase, self).setUp()
        self.controller = notifications.NotificationsController()
        self.req = fakes.HTTPRequest.blank('/v1/notifications',
                                           use_admin_context=True,
                                           version=self.api_version)
        self.context = self.req.environ['masakari.context']

    @mock.patch.object(ha_api.NotificationAPI,
                       'get_notification_recovery_workflow_details')
    def test_show(self, mock_get_notification_recovery_workflow_details):
        (mock_get_notification_recovery_workflow_details
         .return_value) = NOTIFICATION_WITH_PROGRESS_DETAILS

        result = self.controller.show(self.req, uuidsentinel.fake_notification)
        result = result['notification']
        self.assertCountEqual([RECOVERY_OBJ],
                              result.recovery_workflow_details)
        self._assert_notification_data(NOTIFICATION_WITH_PROGRESS_DETAILS,
                                       _make_notification_obj(result))



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Driver base-class:

    (Beginning of) the contract that masakari drivers must follow, and shared
    types that support that contract
"""

import abc
import sys

from oslo_log import log as logging
from stevedore import driver

import masakari.conf
from masakari import utils


CONF = masakari.conf.CONF
LOG = logging.getLogger(__name__)


class NotificationDriver(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def execute_host_failure(self, context, host_name, recovery_method,
                             notification_uuid, **kwargs):
        pass

    @abc.abstractmethod
    def execute_instance_failure(self, context, instance_uuid,
                                 notification_uuid):
        pass

    @abc.abstractmethod
    def execute_process_failure(self, context, process_name, host_name,
                                notification_uuid):
        pass

    @abc.abstractmethod
    def get_notification_recovery_workflow_details(self, context,
                                                   recovery_method,
                                                   notification_uuid):
        pass

    @abc.abstractmethod
    def upgrade_backend(self, backend):
        pass


def load_masakari_driver(masakari_driver=None):
    """Load a masakari driver module.

    Load the masakari driver module specified by the notification_driver
    configuration option or, if supplied, the driver name supplied as an
    argument.

    :param masakari_driver: a masakari driver name to override the config opt
    :returns: a NotificationDriver instance
    """
    if not masakari_driver:
        masakari_driver = CONF.notification_driver

    if not masakari_driver:
        LOG.error("Notification driver option required, but not specified")
        sys.exit(1)

    LOG.info("Loading masakari notification driver '%s'", masakari_driver)
    try:
        notification_driver = driver.DriverManager('masakari.driver',
                                                   masakari_driver,
                                                   invoke_on_load=True).driver
        return utils.check_isinstance(notification_driver, NotificationDriver)
    except ImportError:
        LOG.exception("Failed to load notification driver '%s'.",
                      masakari_driver)
        sys.exit(1)



#!/usr/bin/env python
# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
  CLI interface for masakari management.
"""


import logging as python_logging
import sys
import time

from oslo_config import cfg
from oslo_db.sqlalchemy import migration
from oslo_log import log as logging

import masakari.conf
from masakari import context
from masakari import db
from masakari.db import api as db_api
from masakari.db.sqlalchemy import migration as db_migration
from masakari import exception
from masakari.i18n import _
from masakari import utils
from masakari import version


CONF = masakari.conf.CONF
logging.register_options(CONF)


# Decorators for actions
def args(*args, **kwargs):
    def _decorator(func):
        func.__dict__.setdefault('args', []).insert(0, (args, kwargs))
        return func
    return _decorator


def _db_error(caught_exception):
    print('%s' % caught_exception)
    print(_("The above error may show that the database has not "
            "been created.\nPlease create a database using "
            "'masakari-manage db sync' before running this command."))
    sys.exit(1)


class DbCommands(object):
    """Class for managing the database."""

    def __init__(self):
        pass

    @args('version', nargs='?', default=None, type=int,
          help='Database version')
    def sync(self, version=None):
        """Sync the database up to the most recent version."""
        try:
            return db_migration.db_sync(version)
        except exception.InvalidInput as ex:
            print(ex)
            sys.exit(1)

    def version(self):
        """Print the current database version."""
        print(migration.db_version(db_api.get_engine(),
                                   db_migration.MIGRATE_REPO_PATH,
                                   db_migration.INIT_VERSION))

    @args('--age_in_days', type=int, default=30,
          help='Purge deleted rows older than age in days (default: '
               '%(default)d)')
    @args('--max_rows', type=int, default=-1,
          help='Limit number of records to delete (default: %(default)d)')
    def purge(self, age_in_days, max_rows):
        """Purge rows older than a given age from masakari tables."""
        try:
            max_rows = utils.validate_integer(
                max_rows, 'max_rows', -1, db.MAX_INT)
        except exception.Invalid as exc:
            sys.exit(str(exc))

        try:
            age_in_days = int(age_in_days)
        except ValueError:
            msg = 'Invalid value for age, %(age)s' % {'age': age_in_days}
            sys.exit(str(msg))

        if max_rows == 0:
            sys.exit(_("Must supply value greater than 0 for max_rows."))
        if age_in_days < 0:
            sys.exit(_("Must supply a non-negative value for age."))
        if age_in_days >= (int(time.time()) / 86400):
            sys.exit(_("Maximal age is count of days since epoch."))
        ctx = context.get_admin_context()

        db_api.purge_deleted_rows(ctx, age_in_days, max_rows)


CATEGORIES = {
    'db': DbCommands,
}


def methods_of(obj):
    """Return non-private methods from an object.

    Get all callable methods of an object that don't start with underscore
    :return: a list of tuples of the form (method_name, method)
    """
    result = []
    for i in dir(obj):
        if callable(getattr(obj, i)) and not i.startswith('_'):
            result.append((i, getattr(obj, i)))
    return result


def add_command_parsers(subparsers):
    for category in CATEGORIES:
        command_object = CATEGORIES[category]()

        parser = subparsers.add_parser(category)
        parser.set_defaults(command_object=command_object)

        category_subparsers = parser.add_subparsers(dest='action')

        for (action, action_fn) in methods_of(command_object):
            parser = category_subparsers.add_parser(action)

            action_kwargs = []
            for args, kwargs in getattr(action_fn, 'args', []):
                parser.add_argument(*args, **kwargs)

            parser.set_defaults(action_fn=action_fn)
            parser.set_defaults(action_kwargs=action_kwargs)


command_opt = cfg.SubCommandOpt('category',
                                title='Command categories',
                                help='Available categories',
                                handler=add_command_parsers)


def get_arg_string(args):
    arg = None
    if args[0] == '-':
        # NOTE(Dinesh_Bhor): args starts with FLAGS.oparser.prefix_chars
        # is optional args. Notice that cfg module takes care of
        # actual ArgParser so prefix_chars is always '-'.
        if args[1] == '-':
            # This is long optional arg
            arg = args[2:]
        else:
            arg = args[1:]
    else:
        arg = args

    return arg


def fetch_func_args(func):
    fn_args = []
    for args, kwargs in getattr(func, 'args', []):
        arg = get_arg_string(args[0])
        fn_args.append(getattr(CONF.category, arg))

    return fn_args


def main():
    """Parse options and call the appropriate class/method."""
    CONF.register_cli_opt(command_opt)
    script_name = sys.argv[0]
    if len(sys.argv) < 2:
        print(_("\nOpenStack masakari version: %(version)s\n") %
              {'version': version.version_string()})
        print(script_name + " category action [<args>]")
        print(_("Available categories:"))
        for category in CATEGORIES:
            print(_("\t%s") % category)
        sys.exit(2)

    try:
        CONF(sys.argv[1:], project='masakari',
             version=version.version_string())
        logging.setup(CONF, "masakari")
        python_logging.captureWarnings(True)
    except cfg.ConfigDirNotFoundError as details:
        print(_("Invalid directory: %s") % details)
        sys.exit(2)
    except cfg.ConfigFilesNotFoundError as e:
        cfg_files = ', '.join(e.config_files)
        print(_("Failed to read configuration file(s): %s") % cfg_files)
        sys.exit(2)

    fn = CONF.category.action_fn
    fn_args = fetch_func_args(fn)
    fn(*fn_args)



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import threading

from oslo_config import cfg
from oslo_db import exception as oslo_exception
from oslo_db import options
from stevedore import driver

from masakari import db
from masakari.db import api as db_api
from masakari import exception
from masakari.i18n import _

INIT_VERSION = 0

_IMPL = None
_LOCK = threading.Lock()

options.set_defaults(cfg.CONF)

MIGRATE_REPO_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'migrate_repo',
)


def get_backend():
    global _IMPL
    if _IMPL is None:
        with _LOCK:
            if _IMPL is None:
                _IMPL = driver.DriverManager(
                    "masakari.database.migration_backend",
                    cfg.CONF.database.backend).driver
    return _IMPL


def db_sync(version=None, init_version=INIT_VERSION, engine=None):
    if engine is None:
        engine = db_api.get_engine()

    current_db_version = get_backend().db_version(engine,
                                                  MIGRATE_REPO_PATH,
                                                  init_version)

    if version and int(version) < current_db_version:
        msg = _('Database schema downgrade is not allowed.')
        raise exception.InvalidInput(reason=msg)

    if version and int(version) > db.MAX_INT:
        message = _('Version should be less than or equal to %(max_version)d.'
                    ) % {'max_version': db.MAX_INT}
        raise exception.InvalidInput(reason=message)

    try:
        return get_backend().db_sync(engine=engine,
                                     abs_path=MIGRATE_REPO_PATH,
                                     version=version,
                                     init_version=init_version)
    except oslo_exception.DBMigrationError as exc:
        raise exception.InvalidInput(reason=exc)



# Copyright 2016 NTT DATA.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from masakari.api.validation import parameter_types


_base = {
    'type': 'object',
    'properties': {
        'segment': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name,
                'description': parameter_types.description,
                'recovery_method': {
                    'type': 'string',
                    'enum': ["auto", "reserved_host",
                             "auto_priority", "rh_priority"]
                },
                'service_type': parameter_types.name
            },
            'additionalProperties': False
        }
    },
    'required': ['segment'],
    'additionalProperties': False
}


create = copy.deepcopy(_base)
create['properties']['segment']['required'] = ['name', 'recovery_method',
                                               'service_type']

create_v12 = copy.deepcopy(create)
create_v12['properties']['segment']['properties']['enabled'] = \
    parameter_types.boolean

update = copy.deepcopy(_base)
update['properties']['segment']['anyOf'] = [{'required': ['name']},
                                            {'required': ['description']},
                                            {'required': ['recovery_method']},
                                            {'required': ['service_type']},
                                            ]

update_v12 = copy.deepcopy(update)
update_v12['properties']['segment']['properties']['enabled'] = \
    parameter_types.boolean
update_v12['properties']['segment']['anyOf'].append({'required': ['enabled']})



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from migrate import changeset
from sqlalchemy import Column, MetaData, Table, Index
from sqlalchemy import Integer, DateTime, String, Enum, Text


def define_failover_segments_table(meta):

    failover_segments = Table('failover_segments',
                              meta,
                              Column('created_at', DateTime, nullable=False),
                              Column('updated_at', DateTime),
                              Column('deleted_at', DateTime),
                              Column('deleted', Integer),
                              Column('id', Integer, primary_key=True,
                                     nullable=False),
                              Column('uuid', String(36), nullable=False),
                              Column('name', String(255), nullable=False),
                              Column('service_type', String(255),
                                     nullable=False),
                              Column('description', Text),
                              Column('recovery_method',
                                     Enum('auto', 'reserved_host',
                                          'auto_priority',
                                          'rh_priority',
                                          name='recovery_methods'),
                                     nullable=False),
                              changeset.UniqueConstraint(
                                  'name', 'deleted',
                                  name='uniq_segment0name0deleted'
                              ),
                              changeset.UniqueConstraint(
                                  'uuid',
                                  name='uniq_segments0uuid'),
                              Index('segments_service_type_idx',
                                    'service_type'),
                              mysql_engine='InnoDB',
                              mysql_charset='utf8',
                              extend_existing=True)

    return failover_segments


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    table = define_failover_segments_table(meta)
    table.create()



# Copyright 2016 NTT DATA
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from masakari.cmd import api
from masakari import config
from masakari import exception
from masakari import test


@mock.patch.object(config, 'parse_args', new=lambda *args, **kwargs: None)
class TestMasakariAPI(test.NoDBTestCase):

    def test_continues_without_failure(self):

        fake_server = mock.MagicMock()
        fake_server.workers = 123

        def fake_service(api, **kw):
            return fake_server

        with mock.patch.object(api, 'service') as mock_service:
            mock_service.WSGIService.side_effect = fake_service
            api.main()
            mock_service.WSGIService.assert_has_calls([
                mock.call('masakari_api', use_ssl=False),
            ])
            launcher = mock_service.process_launcher.return_value
            launcher.launch_service.assert_called_once_with(
                fake_server, workers=123)
            self.assertTrue(launcher.wait.called)

    @mock.patch('sys.exit')
    def test_fails_if_all_failed(self, mock_exit):
        mock_exit.side_effect = exception.MasakariException
        with mock.patch.object(api, 'service') as mock_service:
            mock_service.WSGIService.side_effect = exception.PasteAppNotFound(
                name='masakari_api', path='/')
            self.assertRaises(exception.MasakariException, api.main)
            mock_exit.assert_called_once_with(1)
            launcher = mock_service.process_launcher.return_value
            self.assertFalse(launcher.wait.called)
